import glob
import os
import sys
import SCons.Options
import SCons.Environment
import re

def SetupWin32Environment(env, commonLibDir):
	env.Append(CPPPATH = [env['DXSDK_DIR'] + '/Include'])
	env.Append(LIBPATH = [env['DXSDK_DIR'] + '/Lib'])
	env.Append(LIBPATH = [env['DXSDK_DIR'] + '/Lib/x86'])
	
	env.Append(CPPPATH = ['win32/include'])
	env.Append(CPPPATH = [commonLibDir + '/win32/include'])
	env.Append(LIBS = ['kernel32', 'user32', 'dinput8', 'dxguid', 'gdi32', 'Opengl32', 'glu32', 'winmm'])
	
	env.Append(CPPDEFINES = ['COMPILER_MSVC'])
	
	env.Append(CPPDEFINES = ['_CRT_SECURE_NO_DEPRECATE', '_CRT_NONSTDC_NO_DEPRECATE', 'PLAT_WIN32'])
	env.Append(LINKFLAGS = ['/INCREMENTAL:NO'])
	
	if not env['DIST']:
		env.Append(LINKFLAGS = ['/SUBSYSTEM:CONSOLE'])
	else:
		env.Append(LINKFLAGS = ['/SUBSYSTEM:WINDOWS'])
		
	env.Append(CPPFLAGS = ['/EHsc'])
	if env['DEBUG']:
		env.Append(CPPFLAGS = ['/Od', '/Zi', '/WX', '/MDd'])
		env.Append(LINKFLAGS = ['/DEBUG'])
	else:
		env.Append(CPPFLAGS = ['/Ox', '/Ot', '/MD'])
		if env['PROFILE']:
			env.Append(CPPFLAGS = ['/Zi'])
			env.Append(LINKFLAGS = ['/DEBUG'])

	env.Append(ENV = {'INCLUDE' : os.environ['INCLUDE']})
	env.Append(ENV = {'LIB' : os.environ['LIB']})
	
def SetupUnixEnvironment(env, commonLibDir):
	env.Append(LIBS = ['GL', 'GLU', 'pthread'])
	env.Append(LIBS = ['SDL'])

	env.Append(CPPPATH = ['Unix/include'])
	env.Append(CPPPATH = [commonLibDir + '/Unix/include'])

	env.Append(CPPPATH = [env['MAINLIBDIR'] + '/include'])
	env.Append(LIBPATH = [env['MAINLIBDIR'] + '/lib'])

	env.Append(CPPDEFINES = ['PLAT_UNIX'])
	env.Append(CPPDEFINES = ['COMPILER_GCC'])
	
	env.Append(CPPFLAGS = ['-march=pentium4'])
	env.Append(CPPFLAGS = ['-Werror', '-Wall'])
	
	if env['DEBUG']:
		env.Append(CPPFLAGS = ['-g'])
		env.Append(LINKFLAGS = ['-g'])
	else:
		env.Append(CPPFLAGS = ['-O3', '-ffast-math'])
		
	if env['PROFILE']:
		env.Append(CPPFLAGS = ['-p'])
		env.Append(LINKFLAGS = ['-p'])

	if env['NOINLINING']:
		env.Append(CPPFLAGS = ['-fno-inline', '-fno-inline-functions'])
	

def SetupEnvironment(env, commonLibDir):
	env.Append(CPPPATH = [commonLibDir + '/include'])
	env.Append(CPPPATH = ['include'])
	env.Append(LIBPATH = [commonLibDir])
	env.Append(LIBS = ['Common'])

	boostRoot = env['BOOST_ROOT']
	env.Append(CPPPATH = [boostRoot])

	# list of boost libraries that we should link with
	linkBoostLibs = [ 'filesystem', 'system' ]
	
	# Search the compiled boost libraries for the appropriate version
	# and then link with it
	if os.path.exists(boostRoot) and os.path.isdir(boostRoot):
		env.Append(LIBPATH = [boostRoot + '/lib'])
		
		boostLibs = os.listdir(boostRoot + '/lib')
		for boostLib in linkBoostLibs:
			libREString = ''
			if env['DEBUG']:
				libREString = '.*_' + boostLib + '-.*-mt-gd-.*'
			else:
				libREString = '.*_' + boostLib + '-.*-mt-.*'
			
			libRE = re.compile(libREString, re.IGNORECASE)
				
			found = False
			
			for fname in boostLibs:
				# If we find it, link with it!
				if libRE.match(fname) is not None:
					print 'Linking with boost library: ' + fname
					env.Append(LIBS = fname)
					found = True
					break
			
			if not found:
				raise Exception('Could not find boost library -- ' + boostLib + '.  We were looking for it by searching the directory "' + boostRoot + '/lib" for the pattern "' + libREString + '".')
					
	env.Append(CPPDEFINES = ['BOOST_ALL_NO_LIB']) # We just did our own linking, so disable boost's auto-linking		
			
	if env['DEBUG']:
		env.Append(CPPDEFINES = ['DEBUG'])

	# Platform specific setup
	if sys.platform == 'win32' :
		SetupWin32Environment(env, commonLibDir)
	else:
		SetupUnixEnvironment(env, commonLibDir)

def SetupWXWidgets(env):
	# Add all the wxWidgets dependencies
	env.Append(CPPPATH = [env['WXWIDGETS_ROOT']+'/include', env['WXWIDGETS_ROOT']+'/contrib/include'])

	if sys.platform == 'win32' :
		env.Append(CPPPATH = [env['WXWIDGETS_ROOT']+'/include/msvc'])

	env.Append(LIBPATH = [env['WXWIDGETS_ROOT']+'/lib', env['WXWIDGETS_ROOT']+'/contrib/lib'])
	if sys.platform == 'win32' :
		env.Append(LIBPATH = [env['WXWIDGETS_ROOT']+'/lib/vc_lib'])

	if env['DEBUG']:
		env.Append(LIBS = ['wxmsw28d_core', 'wxbase28d'])
	else:
		env.Append(LIBS = ['wxmsw28_core', 'wxbase28'])
		
	env.Append(LIBS = ['winspool', 'comdlg32', 'advapi32', 'shell32', 'ole32', 'oleaut32', 'uuid', 'odbc32', 'odbccp32', 'winmm', 'comctl32', 'rpcrt4', 'wsock32', 'oleacc'])
	
def SetupWXWidgetOptions(opts):
	opts.AddOptions(
		SCons.Options.PathOption('WXWIDGETS_ROOT', 'Path to wxWidgets (what you would set WXWIN to)',	''))


def SetupOptions(opts):	
	opts.AddOptions(
		SCons.Options.BoolOption('DEBUG', 'Build in debug mode', True))
	opts.AddOptions(
		SCons.Options.BoolOption('PROFILE', 'Build with profile instrumentation', False))
	opts.AddOptions(
		SCons.Options.BoolOption('NOINLINING', 'Disable compiler inlining of code', False))
	opts.AddOptions(
		SCons.Options.PathOption('BOOST_ROOT', 'Path to boost headers and libraries',	'/usr/local/boost_1_34_1'))
	opts.AddOptions(
		SCons.Options.BoolOption('DIST', 'Builds the distributable version of the app.  Disables things like console display.', False))
		
	if sys.platform == 'win32':
		opts.AddOptions(
			SCons.Options.PathOption('DXSDK_DIR', 'Path to DirectX SDK', 'C:/Program Files/Microsoft DirectX SDK'))
	else:
		opts.AddOptions(
			SCons.Options.PathOption('MAINLIBDIR', 'Path to SDL directory containing include and lib directories',
			'/usr'))
	
	return opts
	
def MakeCPPEnvironment(opts, commonLibDir):	
	if sys.platform == 'win32':
		toolChain = ['default']
	else:
		toolChain = ['gcc','g++','gnulink','ar']
	
	env = SCons.Environment.Environment(options = opts, tools = toolChain)
	
	env.Help(opts.GenerateHelpText(env))

	SetupEnvironment(env, commonLibDir)
	opts.Save('options.cache', env)
	
	return env


def MakeMSVCProject(env, ProjectName, SourceFiles, IncludeFiles):
	# Build an MSVS project
	projectNode = env.MSVSProject(target = ProjectName + env['MSVSPROJECTSUFFIX'],
								srcs = SourceFiles,
								incs = IncludeFiles,
								auto_build_solution=0,
#								buildtarget = ProjectName,
								runfile = ProjectName + '/' + ProjectName,
								variant = 'Release')

	# Finally make the solution containing the projects just created
	solutionNode = env.MSVSSolution(target = ProjectName + env['MSVSSOLUTIONSUFFIX'],
									projects = [projectNode],
									variant = 'Release')

	env.Depends(solutionNode, projectNode) # It'd be nice if MSVSSolution did this automatically

	return solutionNode

