#include <OSWindow.h>

#include "XWindowGraphicsContext.h"
#include "XWindowKeyboardInput.h"
#include "XWindowMouseInput.h"

#include <X11/Xlib.h>
#include <GL/glx.h>

#include "DefaultDisplay.h"

#include <ThreadPrimitives.h>
#include <boost/bind.hpp>

#include <iostream>

static int MyXErrorHandler(Display* disp, XErrorEvent* error)
{
	char errString[500];
	XGetErrorText(disp, error->error_code, errString, 500);
	std::cout << "X Windows Error: " << errString << std::endl;
	return 0;
}

static Bool IsThisWindow(Display* disp, XEvent* event, XPointer arg)
{
	Window* w = (Window*)arg;
	return event->xany.window == *w;
}

class XWindow_Pimpl : public OSWindow::Pimpl
{
	Mutex m_monitorLock;

	Thread* m_winThread;

	bool m_wantsToQuit;
	Condition m_quitCondition;

	XWindowKeyboardInput m_keyboardInput;
	XWindowMouseInput m_mouseInput;

	XWindowGraphicsContext* m_graphicsContext;

	Window m_win;
	Display* m_disp;

	Atom m_WM_DELETE_WINDOW;
	int m_width;
	int m_height;
	std::string m_title;

	void MarkToQuit()
	{
		Mutex::ScopedLock lock(m_monitorLock);
	
		m_wantsToQuit = true;
		m_quitCondition.Signal();
	}

	Condition m_initDoneCond;

	void Run()
	{
		m_disp = GetDefaultDisplay();
		ASSERT(m_disp != NULL);

		int defScreen = DefaultScreen(m_disp);


		static int attrList[] = { 
//			GLX_RGBA,
			GLX_RED_SIZE, 8, 
			GLX_GREEN_SIZE, 8, 
			GLX_BLUE_SIZE, 8,

			GLX_DOUBLEBUFFER, True,

			GLX_DEPTH_SIZE, 16,

			None 
		};


		int numConfigs;
		GLXFBConfig* configs = glXChooseFBConfig(m_disp, defScreen, attrList, &numConfigs);
		ASSERT(configs != NULL);

		XSetWindowAttributes attrs;
		attrs.border_pixel = 0;
		attrs.event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | PointerMotionMask | StructureNotifyMask;
		
		m_win = XCreateWindow(	m_disp, RootWindow(m_disp, defScreen), 
						0, 0,
						m_width, m_height, 2, DefaultDepth(m_disp, defScreen),
						InputOutput,
						glXGetVisualFromFBConfig(m_disp, configs[0])->visual,
						CWBorderPixel | CWEventMask,
						&attrs);

		XSetStandardProperties(m_disp, m_win, m_title.c_str(), m_title.c_str(), None, NULL, 0, NULL);

		m_graphicsContext = new XWindowGraphicsContext(m_win, configs[0], m_width, m_height);

		m_WM_DELETE_WINDOW = XInternAtom(m_disp, "WM_DELETE_WINDOW", 0);
		XSetWMProtocols(m_disp, m_win, &m_WM_DELETE_WINDOW, 1);

		XMapWindow(m_disp, m_win);

		XFlush(m_disp);

		{
			Mutex::ScopedLock lock(m_monitorLock);
			m_initDoneCond.Signal();
		}

		// Enter event loop
		while(!m_wantsToQuit)
		{
			XEvent xevent;
			XIfEvent(m_disp, &xevent, IsThisWindow, (XPointer)&m_win);

			switch(xevent.type)
			{
				case ButtonPress:
				{
					m_mouseInput.PostOSButtonEvent(xevent.xbutton.button, true);
				} break;
				case ButtonRelease:
				{
					m_mouseInput.PostOSButtonEvent(xevent.xbutton.button, false);
				} break;
				case MotionNotify:
				{
					m_mouseInput.PostOSMouseMoveEvent(xevent.xmotion.x, xevent.xmotion.y);
				} break;
				case KeyPress:
				{
					m_keyboardInput.PostOSKeyEvent(xevent.xkey.keycode, true);
				} break;	
				case KeyRelease:
				{
					m_keyboardInput.PostOSKeyEvent(xevent.xkey.keycode, false);
				} break;	
				case ClientMessage:
				{
					if((xevent.xclient.format == 32)
						&& ((Atom)xevent.xclient.data.l[0] == m_WM_DELETE_WINDOW))
					{
						MarkToQuit();
					}
				} break;
			};
		}
	}

	public:
	XWindow_Pimpl(const std::string& winTitle, int width, int height)
		: m_wantsToQuit(false)
		, m_graphicsContext(0)
		, m_width(width)
		, m_height(height)
		, m_title(winTitle)
	{
		XSetErrorHandler(&MyXErrorHandler);

		Mutex::ScopedLock lock(m_monitorLock);

		m_winThread = new Thread(boost::bind(&XWindow_Pimpl::Run, this));

		m_initDoneCond.Wait(m_monitorLock);

		m_graphicsContext->SetCurrentContext();
		m_graphicsContext->ResizeScreen(width, height);
	}
	
	~XWindow_Pimpl()
	{
		// Send the quit event to the window, close down time
		XEvent quitEvent;
		quitEvent.type = ClientMessage;
		quitEvent.xclient.format = 32;
		quitEvent.xclient.data.l[0] = m_WM_DELETE_WINDOW;
		quitEvent.xany.window = m_win;
		XSendEvent(m_disp, m_win, True, 0, &quitEvent);
		XFlush(m_disp);

		delete m_winThread;
		delete m_graphicsContext;
		XDestroyWindow(m_disp, m_win);
		XFlush(m_disp);
	}

	GraphicsContext& GetGraphicsContext()
	{
		ASSERT(m_graphicsContext != 0);
		return *m_graphicsContext;
	}

	KeyboardInput& GetKeyboardInput()
	{
		return m_keyboardInput;
	}

	MouseInput& GetMouseInput()
	{
		return m_mouseInput;
	}

	void SetWindowMode(OSWindow::Mode mode)
	{
	}

	OSWindow::Mode GetWindowMode()
	{
		return OSWindow::M_WINDOWED;
	}

	bool WantsToQuit() const
	{
		return m_wantsToQuit;
	}

	void WaitForQuit()
	{
		Mutex::ScopedLock lock(m_monitorLock);
	
		while(!m_wantsToQuit)
		{
			m_quitCondition.Wait(m_monitorLock);
		}
	}
};

OSWindow::OSWindow(const std::string& winTitle, int width, int height)
: m_pimpl(new XWindow_Pimpl(winTitle, width, height))
{
}

OSWindow::~OSWindow()
{
	delete m_pimpl;
}

