#ifndef COMMANDQUEUE_H_
#define COMMANDQUEUE_H_

#include <FIFOAllocator.h>
#include <BoundedQueue.h>
#include <ThreadPrimitives.h>

#include <boost/bind.hpp>

class CommandQueue
{
public:
	class Command
	{
	public:
		virtual ~Command() {};

		virtual void ProcessCommand() = 0;
	};


private:
	friend void* operator new(unsigned int num_bytes, CommandQueue& cQueue);
	friend void operator delete(void* mem, CommandQueue& cQueue);

	FIFOAllocator m_commandMemory;
	BoundedQueue<Command*> m_commandQueue;

	Thread* m_commandProcessingThread;

	Mutex m_processingMutex;
	Condition m_doneProcessingCond;

	Mutex m_addCommandMutex;

public:
	CommandQueue(int maxQueueItems, int maxQueueMem, bool createThread=true)
		: m_commandMemory(maxQueueMem, true)
		, m_commandQueue(maxQueueItems)
		, m_commandProcessingThread(0)
	{
		if(createThread)
		{
			m_commandProcessingThread = new Thread(boost::bind<void>(&CommandQueue::ProcessCommands, this));
		}
	};
	virtual ~CommandQueue()
	{
		Mutex::ScopedLock lock(m_processingMutex);

		m_commandQueue.Push(0); // Quit command

		m_doneProcessingCond.Wait(m_processingMutex);

		if(m_commandProcessingThread != 0)
		{
			delete m_commandProcessingThread;
		}
	}

	// These functions shouldn't be called directly,
	// use the ADDCOMMAND macro
	void PreAddCommand()
	{
		m_addCommandMutex.Lock();
	}
	void PostAddCommand()
	{
		m_addCommandMutex.Unlock();
	}
	void AddCommand(Command* command)
	{
		ASSERT(m_commandMemory.Own(command));
		m_commandQueue.Push(command);
	}

	void Flush()
	{
		m_commandQueue.WaitEmpty();
	}

	// Called internally if no thread is passed in, otherwise you can call
	// this from your own thread to begin the command processor
	void ProcessCommands()
	{
		while(true)
		{
			Command* nextCommand = m_commandQueue.Pop();

			if(nextCommand == 0) break;

			nextCommand->ProcessCommand();

			nextCommand->~Command();
			m_commandMemory.Free(nextCommand);
		}

		ASSERT(m_commandMemory.Empty());

		Mutex::ScopedLock lock(m_processingMutex);
		m_doneProcessingCond.SignalAll();
	}
};

inline void* operator new(unsigned int num_bytes, CommandQueue& cQueue)
{
	return cQueue.m_commandMemory.Allocate(num_bytes);
}

inline void operator delete(void* mem, CommandQueue& cQueue)
{
	return cQueue.m_commandMemory.Free(mem);
}

class CommandHandler
{
public:
	virtual CommandQueue& GetCommandQueue() = 0;
};

class Command : public CommandQueue::Command
{
public:
	virtual void ProcessCommand() {Process();}
	virtual void Process() = 0;
};

#define HANDLES(x) public x##Handler

#define DECL_COMMAND(x) \
	struct x; \
	class x##Handler : virtual public CommandHandler\
	{ \
	public: \
		virtual void ProcessCommand(x&) = 0; \
	}; \
	\
	struct x : public Command

#define COMMAND_BODY(x) \
	x##Handler* m_handler; \
	x* SetHandler(x##Handler& pHandler) {m_handler = &pHandler; return this;} \
	virtual void Process() {m_handler->ProcessCommand(*this);}

#define ADDCOMMAND(q, c) \
	do { \
		(q).PreAddCommand(); \
		(q).AddCommand(new(q) c); \
		(q).PostAddCommand(); \
	} while(false)

#define ADDCOMMAND2(q, c) \
	do { \
		(q).GetCommandQueue().PreAddCommand(); \
		(q).GetCommandQueue().AddCommand((new(q.GetCommandQueue()) c)->SetHandler(q)); \
		(q).GetCommandQueue().PostAddCommand(); \
	} while(false)

#endif
