#ifndef FILEFETCHER_H_
#define FILEFETCHER_H_

#include <ThreadPrimitives.h>
#include <ResourceIdentifier.h>
#include <BulkFileIO.h>

#include <set>
#include <vector>
#include <list>


// The FileFetcher acts as a system to load files for you ASAP.
// You submit a file load request by creating an instance of the FileFetcher::FileLoad class,
// and then you can either query whether the load is complete by calling FileFetcher::FileLoad::Done()
// or by passing in a callback function to call when the load is complete.  Beware, the callback will
// be called from a seperate thread, with a lock (so don't call back in to the FileFetcher). 
// You may destroy the FileFetcher::FileLoad object whenever you want, even if the load is not
// finished.  If the FileFetcher::FileLoad object is destroyed before the load is finished, the
// callback will not be called.

class FileFetcherLoadWorker;
class FileFetcher
{
public:
	struct WorkUnit;

	// This FileLoad class is basically the public interface to the FileFetcher.
	// By constructing a FileLoad, you are requesting that the FileFetcher fetch the given
	// file for you.  If you destroy the FileLoad, then the work order is lost.
	class FileLoad
	{
		friend class FileFetcher;

		FileFetcher& m_fileFetcher;
		ResourceIdentifier m_resId;

		mutable Mutex m_monitorMutex;

		void* m_data;
		long m_dataSize;
		bool m_error;

		WorkUnit* m_currentWorkUnit; // 0 until a worker is assigned to this file load
		std::list<FileLoad*>::iterator m_positionInLine; // Iterator pointing to this object in to the wait queue

		typedef void FileLoadDoneSig(FileLoad*);
		boost::function<FileLoadDoneSig> m_doneCB;

		bool PrivateIsDone() const {return m_data != 0 || m_error;}

	public:

		FileLoad(FileFetcher& fetcher, const ResourceIdentifier& resId, const boost::function<FileLoadDoneSig>& doneCB = boost::function<FileLoadDoneSig>());
		~FileLoad();

		bool Done() const;
		bool Error() const;

		const void* GetData() const;
		long GetDataSize() const;

		void StopLoadProcess(); // Stops loading this file (ie: callback will never get called).  Does nothing if the file is already loaded

		const ResourceIdentifier& GetResId() const {return m_resId;}
	};

private:
	FileFetcher(const FileFetcher&);
	const FileFetcher& operator=(const FileFetcher&);

	BulkFileIO::FileMemAllocator m_fetcherAllocator; // The platform specific allocator that we pass in to the workers to use to allocate space to put the files in

	std::vector<FileFetcherLoadWorker*> m_workers; // The file loader workers that we're managing

	std::list<FileLoad*> m_queuedFileLoads; // This is the list of file loads that we had no free workers available to tend to at the request time
	Condition m_fileLoadQueuedCond;

	mutable Mutex m_monitorMutex;

	void RegisterFileLoad(FileLoad* loadToRegister);
	void UnregisterFileLoad(FileLoad* loadToUnregister);
	void StopLoadProcess(FileLoad* loadToStop);
	void PrivateStopLoadProcess(FileLoad* loadToStop);

	int m_workersToShutdown;
	Condition m_allWorkersShutdownCond;

	void StartupWorkers(); // Initialize our worker threads
	void ShutdownWorkers(); // Waits for the worker threads to all shutdown

public:
	FileFetcher(long int maxMem);
	virtual ~FileFetcher();

	BulkFileIO::FileMemAllocator& GetFileAllocator() {return m_fetcherAllocator;}

	WorkUnit* RequestWork(); // Workers call this and the FileFetcher gives them a file to load
	void ReturnWorkResults(WorkUnit* fileLoad); // Workers call this when they're done loading a file

	// This is what we hand off to the workers.  They fill out the data and dataSize, and then
	// hand it back to us.
	struct WorkUnit
	{
		FileLoad* fileLoad;
		ResourceIdentifier resId;
		void* data;
		long dataSize;

		WorkUnit(FileLoad* pFileLoad)
			: fileLoad(pFileLoad)
			, resId(fileLoad->GetResId())
			, data(0)
			, dataSize(-1)
		{};

	private:
		WorkUnit(const WorkUnit& other);
		const WorkUnit& operator=(const WorkUnit& other);
	};
};

#endif
