#ifndef OBJFILE_H_
#define OBJFILE_H_

#include <SafePointerList.h>

#include <Algebra.h>
#include <vector>
#include <map>

class ResourceManager;
class OBJFile
{
public:
	typedef Vector<3, VT_INSTANCE, false> Vertex;
	typedef std::vector< Vertex > VertexList;

	struct Material
	{
		std::string name;
		float shininess;
		Vector<3, VT_INSTANCE, false> ambientColour;
		Vector<3, VT_INSTANCE, false> diffuseColour;
		Vector<3, VT_INSTANCE, false> specularColour;
		float alpha;

		bool useSpecular;

		Material()
			: useSpecular(false)
		{};
	};

	struct Face
	{
		Face()
			: numVertices(0)
		{
			for(int i=0; i<4; ++i) vertexIndices[i] = -1;
			for(int i=0; i<4; ++i) normalIndices[i] = -1;
		}
		int numVertices;
		int vertexIndices[4];
		int normalIndices[4];

		const Material* material;
	};
	typedef std::vector< Face > FaceList;

	struct Object
	{
	
		std::string name;

		VertexList vertices;
		VertexList normals;

		FaceList faces;

		int GetNumVertices() const {return vertices.size();}
		const Vertex& GetVertex(int index) const {return vertices[index];}
		
	};

private:

	SafePointerList<Object> m_objectList;
	SafePointerList<Material> m_materialList;

	typedef std::map<std::string, Object*> ObjectMap;
	ObjectMap m_objects;

	typedef std::map<std::string, Material*> MaterialMap;
	MaterialMap m_materials;

public:
	OBJFile(std::istream& in, ResourceManager& resMan); // Path is needed since the OBJ file might reference other MTL files
	virtual ~OBJFile();

	int GetNumObjects() const {return m_objectList.objs.size();}
	const Object& GetObjectItem(int index) const {return *m_objectList.objs[index];}
	const Object& GetObjectItem(const std::string& objName) const
	{
		ObjectMap::const_iterator found = m_objects.find(objName);
		ASSERT(found != m_objects.end());
		return *found->second;
	}

	int GetNumMaterials() const {return m_materialList.objs.size();}
	const Material* GetMaterial(int index) const {return m_materialList.objs[index];}
	const Material* GetMaterial(const std::string& objName) const
	{
		MaterialMap::const_iterator found = m_materials.find(objName);
		ASSERT(found != m_materials.end());
		return found->second;
	}
};

#endif
