#ifndef POTENTIALOBJECT_H_
#define POTENTIALOBJECT_H_

///
/// \brief A placeholder for an instance of the given class.  Will reserve the
///			memory for the object, but won't construct it automatically.
///
///		The PotentialObject class can help you out if you're looking to take advantage
///		of RAII, but do not wish to take the non constant-time allocation hit.  The
///		PotentialObject will reserve as a member the memory needed for the specified class.
///

template <class T>
class PotentialObject
{
	int dummyMemberToEncourageAlignment;

	char m_objMem[sizeof(T)];
	bool m_exists;

	template <class F> friend void* operator new(unsigned int num_bytes, PotentialObject<F>& objSlot);
	template <class F> friend void operator delete(void* mem, PotentialObject<F>& objSlot);

public:
	PotentialObject()
		: m_exists(false)
	{};
	virtual ~PotentialObject()
	{
		if(Exists()) Clear();
	}

	PotentialObject(const PotentialObject<T>& other)
	: m_exists(false)
	{
		operator=(other);
	}
	const PotentialObject<T>& operator=(const PotentialObject<T>& other)
	{
		if(&other != this)
		{
			if(Exists()) Clear();
			if(other.m_exists)
			{
				new(*this) T(*other);
			}
		}
		return *this;
	}
	
	bool Exists() const {return m_exists;}

	T* GetPtr() {ASSERT(Exists()); return reinterpret_cast<T*>(m_objMem);}
	const T* GetPtr() const {ASSERT(Exists()); return reinterpret_cast<const T*>(m_objMem);}

	T* operator->() {return GetPtr();}
	T& operator*() {return *GetPtr();}

	const T* operator->() const {return GetPtr();}
	const T& operator*() const {return *GetPtr();}

	void Clear()
	{
		ASSERT(Exists());
		reinterpret_cast<T*>(m_objMem)->~T();
		m_exists = false;
	} 
};

template <class T>
void* operator new(unsigned int num_bytes, PotentialObject<T>& objSlot)
{
	ASSERT(num_bytes == sizeof(T));
	ASSERT(!objSlot.Exists());

	objSlot.m_exists = true;

	return reinterpret_cast<T*>(objSlot.m_objMem);
}

template <class T>
void operator delete(void* mem, PotentialObject<T>& objSlot)
{
	ASSERT(objSlot.Exists());
	ASSERT(mem == objSlot.m_objMem);

	objSlot.m_exists = false;
}


#endif

