#ifndef RESOURCEMANAGER_H_
#define RESOURCEMANAGER_H_

#include <ResourceCreator.h>
#include <Allocator.h>
#include <ThreadPrimitives.h>

#include <PotentialObject.h>

#include <map>
#include <list>

// The PriorityAllocator is responsible for allowing resource allocations up to a maximum
// capacity.  Once this is reached, it will start automatically freeing the blocks marked
// freeable.  When a block is about to be freed, a callback will be fired (destruct your object).
// When an allocation is made, it is by default marked as unfreeable.  Once it is marked
// as freeable, it may get freed to make room for other allocations.
// Allocations are actually made by the passed in allocator, so this class is a decorator.

class ResourceManager : public ThreadSafeAllocator
{
public:
	typedef void FreeingDataSig(void*);

private:
	Allocator& m_alloc;
	const long m_maxCapacity;

	ResourceCreator* m_resCreator;

	struct ResourceAcquisition;
	typedef std::map<ResourceIdentifier, ResourceAcquisition*> AcquisitionMap;
	struct AllocationInfo;
	typedef std::map<void*, AllocationInfo*> AllocationMap;

	typedef std::list<AllocationMap::iterator> FreeableList;

	struct AllocationInfo
	{
		long allocationSize;

		FreeableList::iterator freeablePosition; // Our position in the freeable list (the end, if we're not freeable)

		AcquisitionMap::iterator resAcquisition; // The link to the acquisition information
	};
	AllocationMap m_allocations; // Tracks ALL allocations

	long m_curMemUsed;
	long m_curNonFreeableMemUsed;

	FreeableList m_freeableList; // This is used as a queue to determine what to free first


	struct ResourceAcquisition
	{
		std::auto_ptr<ResourceFactory> resFactory;

		int acquisitionCount; // How many times has this been acquired?

		PotentialObject<ResourceCreator::CreationProcess> creationProcess;

		bool error;

		mutable Condition resourceAvailableCond; // Signaled when the resource becomes available (the link is made to the allocation)
		AllocationMap::iterator allocInfo; // The link to the allocation information
	};
	AcquisitionMap m_acquisitions;

	Thread* m_resCreatedHandlerThread;
	struct ResCreatedInfo
	{
		ResourceCreator::CreationProcess* creationProcess;
		void* data;
	};
	std::list<ResCreatedInfo*> m_resCreatedQueue;
	Mutex m_resCreatedMutex;
	Condition m_resCreatedCond;
	Condition m_noResToCreateCond;
	void RunResCreatedHandlingThread();
	void ProcessResourceCreated(ResourceCreator::CreationProcess* creatingProcess, void* data); // This function actually handles the resource creation, in another thread
	void HandleResourceCreated(ResourceCreator::CreationProcess* creatingProcess, void* data); // This function only queues the parameters

	Thread* m_resDestroyerThread;
	void RunResourceDestroyer();
	struct ResToDestroyInfo
	{
		void* data;
		std::auto_ptr<ResourceFactory> resFactory;
	};
	std::list<ResToDestroyInfo*> m_destroyList;
	Mutex m_resToDestroyMutex;
	Condition m_noResToDestroyCond; // Signaled when the m_destroyList is empty
	Condition m_resToDestroyCond;


	mutable Mutex m_monitorMutex;


	void SetFreeableFlag(AllocationMap::iterator& allocIter, bool freeable);

	void Free(AllocationMap::iterator iter);

	// Called through ResourceReference
	template <class T, class FACTORY> 
	friend class ResourceReference;

	AcquisitionMap::iterator Acquire(const ResourceIdentifier& resId, std::auto_ptr<ResourceFactory> resFactory);
	AcquisitionMap::iterator Reacquire(const AcquisitionMap::iterator& iter);
	void Release(AcquisitionMap::iterator& iter);

	enum AcquisitionStatus
	{
		ACQUIRING,
		ACQUIRED,
	};
	AcquisitionStatus GetAcquisitionStatus(const AcquisitionMap::iterator& iter) const;
	void* GetAcquiredResourceData(const AcquisitionMap::iterator& iter) const; // Returns the actual resource data, waiting if it's not currently available

public:
	ResourceManager(Allocator& alloc, long maxCapacity, FileFetcher& fetcher);
	~ResourceManager();

	virtual void* Allocate(long bytes);
	virtual void Free(void* mem);
};


template <class T>
class DefaultFactory : public TypedResourceFactory<T>
{
	ResourceManager& m_resMan;

public:
	DefaultFactory(ResourceManager& resMan)
		: m_resMan(resMan)
	{};

	virtual void CreateObject(void* mem, std::istream& in)
	{
		new(mem) T(in, m_resMan);
	}
};

template <class T, class FACTORY=DefaultFactory<T> >
class ResourceReference
{
	ResourceManager& m_resAlloc;
	ResourceManager::AcquisitionMap::iterator m_acquired;

public:
	ResourceReference(ResourceManager& resAlloc, const ResourceIdentifier& resId)
		: m_resAlloc(resAlloc)
	{
		m_acquired = m_resAlloc.Acquire(resId, std::auto_ptr<ResourceFactory>(new FACTORY(m_resAlloc)));
	}
	ResourceReference(const ResourceReference& other)
		: m_resAlloc(other.m_resAlloc)
	{
		m_acquired = m_resAlloc.Reacquire(other.m_acquired);
	}
	const ResourceReference& operator=(const ResourceReference& other)
	{
		if(this != &other)
		{
			m_resAlloc.Release(m_acquired);
			m_acquired = m_resAlloc.Reacquire(other.m_acquired);
		}

		return *this;
	}
	~ResourceReference()
	{
		m_resAlloc.Release(m_acquired);
	}

	// Returns true if the resource has finished loading
	bool IsLoaded() const
	{
		return m_resAlloc.GetAcquisitionStatus(m_acquired) == ResourceManager::ACQUIRED;
	}

	void WaitUntilLoaded() const
	{
		m_resAlloc.GetAcquiredResourceData(m_acquired);
	}

	// Blocks if it's not available yet, use IsLoaded() to check
	const T& operator*() const
	{
		ASSERTMSG(m_resAlloc.GetAcquiredResourceData(m_acquired) != 0, "Failed to get this resource!");
		return *((T*)m_resAlloc.GetAcquiredResourceData(m_acquired));
	}
	const T* operator->() const
	{
		ASSERTMSG(m_resAlloc.GetAcquiredResourceData(m_acquired) != 0, "Failed to get this resource!");
		return (T*)m_resAlloc.GetAcquiredResourceData(m_acquired);
	}
	const T* GetPtr() const
	{
		ASSERTMSG(m_resAlloc.GetAcquiredResourceData(m_acquired) != 0, "Failed to get this resource!");
		return (T*)m_resAlloc.GetAcquiredResourceData(m_acquired);
	}
};

#endif
