#include "Model.h"
#include "OBJFile.h"
#include <GraphicsFrame.h>

Model::Model(const OBJFile::Object& objFileObject)
{
	// This map will link materials we've already seen to the meshes we have constructed for them
	// This helps us to quickly decide which mesh to add faces to, based on the face material
	typedef std::map<Material*, MaterialMesh*> MaterialMeshMap;
	MaterialMeshMap materialToMeshMap;

	typedef std::map<int, int> VertMap; // Used later to map OBJFile vertex indices to mesh vertex indices
	typedef std::map<MaterialMesh*, VertMap> MeshVertMapMap;
	MeshVertMapMap meshVertMapMap;

	typedef std::map<const OBJFile::Material*, Material*> OBJMatToModelMat;
	OBJMatToModelMat toModelMat;

	// Go through and add all the faces
	const OBJFile::FaceList& faceList = objFileObject.faces;
	for(int i=0; i<faceList.size(); ++i)
	{
		const OBJFile::Face& curFace = faceList[i];

		// First, find the standard material based on the OBJ file material
		// if it doesn't exist, create it (ie: we're transferring the OBJ file
		// materials in to Material objects stored within the Model object)
		Material* standardMat;
		OBJMatToModelMat::const_iterator found = toModelMat.find(curFace.material);
		if(found == toModelMat.end())
		{
			ASSERTMSG(curFace.material != 0, "All model faces must have an assigned material.");

			// Add it.
			standardMat = new Material(*curFace.material);
			m_materials.objs.push_back(standardMat);
			toModelMat[curFace.material] = standardMat;
		}
		else
		{
			standardMat = found->second;
		}

		// We need to decide which mesh to add the face to, based on the material
		MaterialMesh* addMesh = 0;
		
		MaterialMeshMap::const_iterator matMeshFound = materialToMeshMap.find(standardMat);
		if(matMeshFound == materialToMeshMap.end())
		{
			// Create a new mesh, add it to our mesh list and the material->mesh map
			m_materialMeshList.push_back(MaterialMesh());
			addMesh = &m_materialMeshList.back();
			addMesh->material = standardMat;
			materialToMeshMap[standardMat] = addMesh;
			meshVertMapMap[addMesh] = VertMap(); // Give it a fresh vert map
		}
		else
		{
			addMesh = matMeshFound->second;
		}

		ASSERT(addMesh != 0);

		// Get the vertex map for this mesh
		MeshVertMapMap::iterator foundVertMap = meshVertMapMap.find(addMesh);
		ASSERT(foundVertMap != meshVertMapMap.end());
		VertMap& curVertMap = foundVertMap->second;

		// Okay, now add this face to the mesh.  First we add the vertices of the triangle to the mesh, if they don't already exist
		for(int i=0; i<curFace.numVertices; ++i)
		{
			int vertInd = -1;
			VertMap::const_iterator foundVert = curVertMap.find(curFace.vertexIndices[i]);
			if(foundVert == curVertMap.end())
			{
				// add it
				vertInd = addMesh->mesh.AddVertex(objFileObject.vertices[curFace.vertexIndices[i]]);
				curVertMap[curFace.vertexIndices[i]] = vertInd;
			}
			else
			{
				vertInd = foundVert->second;
			}
			ASSERT(vertInd != -1);
		}

		// Okay, we can be certain now that the mesh contains all needed vertices for the triangle
		if(curFace.numVertices == 3)
		{
			TriangleMesh::Triangle tri(curVertMap[curFace.vertexIndices[0]], curVertMap[curFace.vertexIndices[1]], curVertMap[curFace.vertexIndices[2]]);

			if(curFace.normalIndices[0] != -1 && curFace.normalIndices[1] != -1 && curFace.normalIndices[2] != -1)
			{
				tri.SetNormals(objFileObject.normals[curFace.normalIndices[0]], objFileObject.normals[curFace.normalIndices[1]], objFileObject.normals[curFace.normalIndices[2]]);
			}

			addMesh->mesh.AddTriangle(tri);
		}
		else if(curFace.numVertices == 4)
		{
			// Add the first triangle of the quad
			TriangleMesh::Triangle tri1(curVertMap[curFace.vertexIndices[0]], curVertMap[curFace.vertexIndices[1]], curVertMap[curFace.vertexIndices[2]]);

			if(curFace.normalIndices[0] != -1 && curFace.normalIndices[1] != -1 && curFace.normalIndices[2] != -1)
			{
				tri1.SetNormals(objFileObject.normals[curFace.normalIndices[0]], objFileObject.normals[curFace.normalIndices[1]], objFileObject.normals[curFace.normalIndices[2]]);
			}

			addMesh->mesh.AddTriangle(tri1);

			// and now the second
			TriangleMesh::Triangle tri2(curVertMap[curFace.vertexIndices[3]], curVertMap[curFace.vertexIndices[0]], curVertMap[curFace.vertexIndices[2]]);

			if(curFace.normalIndices[3] != -1 && curFace.normalIndices[0] != -1 && curFace.normalIndices[2] != -1)
			{
				tri2.SetNormals(objFileObject.normals[curFace.normalIndices[3]], objFileObject.normals[curFace.normalIndices[0]], objFileObject.normals[curFace.normalIndices[2]]);
			}

			addMesh->mesh.AddTriangle(tri2);
		}
		else
		{
			ASSERTMSG(false, "Unsupported number of vertices");
		}
	}

	// Now setup any normals that weren't set
	for(MaterialMeshList::iterator iter = m_materialMeshList.begin(); iter != m_materialMeshList.end(); ++iter)
	{
		iter->mesh.FillInNormals();
	}
}

void Model::AddToGraphicsFrame(GraphicsFrame& frame, const Matrix<4,4>& modelView, float alpha, int layer, bool wireframe) const
{
	int matIndex = frame.RegisterMatrix(modelView);

	for(MaterialMeshList::const_iterator iter = m_materialMeshList.begin();
		iter != m_materialMeshList.end();
		++iter)
	{
		frame.AddPacket(GraphicsFrame::Packet(matIndex, &iter->mesh, iter->material, alpha, layer, wireframe));
	}
}
