#include "SDLWindow.h"

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include <Utils.h>

#include <sstream>


SDLWindowGraphicsContext::SDLWindowGraphicsContext(int width, int height)
: m_viewportWidth(width)
, m_viewportHeight(height)
, m_near(0.02f)
, m_far(500.0f)
{
}

SDLWindowGraphicsContext::~SDLWindowGraphicsContext()
{
}

void SDLWindowGraphicsContext::ResizeScreen(int width, int height)
{
	m_viewportWidth = width;
	m_viewportHeight = height;
}

void SDLWindowGraphicsContext::RenderGraphicsFrame(GraphicsFrame& frame)
{
	GraphicsFrame::Freeze freeze = frame.MakeFreeze();

	m_glContext.RenderGraphicsFrame(frame, GetProjectionMatrix());
	SDL_GL_SwapBuffers();
}


Matrix<4,4> SDLWindowGraphicsContext::GetProjectionMatrix()
{
	return MakeProjectionMatrix(45.0f, (float)m_viewportWidth/m_viewportHeight, m_near, m_far);
}

Matrix<4,4> SDLWindowGraphicsContext::MakeProjectionMatrix(float fov, float aspectRatio, float n, float f) const
{
	float fovRad = 2*M_PI * fov / 360.0f; 

	float t = n * tan(fovRad/2.0f);
	float b = -t;
	float r = t * aspectRatio;
	float l = -r;

	Matrix<4,4> proj(Matrix<4,4>::Zero);
	proj.GetVal(0,0) = 2*n / (r - l);
	proj.GetVal(1,1) = 2*n / (t - b);
	proj.GetVal(0,2) = (r+l) / (r-l);
	proj.GetVal(1,2) = (t+b) / (t-b);
	proj.GetVal(2,2) = - (f + n) / (f - n);
	proj.GetVal(3,2) = -1;
	proj.GetVal(2,3) = - (2 * f * n) / (f - n);

	return proj;
}


void SDLWindowMouseInput::PollMouse()
{
	int mouseX, mouseY;
	SDL_GetRelativeMouseState(&mouseX, &mouseY);

	MoveEvent moveEvent;
	moveEvent.x = mouseX;
	moveEvent.y = mouseY;

	PostMoveEvent(moveEvent);
}


SDLWindow::SDLWindow(const std::string& title, int width, int height)
: m_title(title)
, m_width(width)
, m_height(height)
{
	if(SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		ASSERT(false);
		std::ostringstream ss;
		ss << "SDL_Init failed: " << SDL_GetError();
		throw Exception(Exception::EID_FATAL, ss.str());
	}

	SDL_WM_SetCaption(title.c_str(), 0);

	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	int flags = SDL_HWSURFACE | SDL_OPENGL;

	m_screen = SDL_SetVideoMode(width, height, 32, flags);
	if(m_screen == NULL)
	{
		SDL_QuitSubSystem(SDL_INIT_VIDEO);

		ASSERT(false);
		std::ostringstream ss;
		ss << "SDL_SetVideoMode failed: " << SDL_GetError();
		throw Exception(Exception::EID_FATAL, ss.str());
	}

	m_graphicsContext = new SDLWindowGraphicsContext(width, height);
	m_graphicsContext->ResizeScreen(width, height);
}

SDLWindow::~SDLWindow()
{
	delete m_graphicsContext;
}

void SDLWindow::PumpEvents()
{
	SDL_Event curEvent;
	while(SDL_PollEvent(&curEvent))
	{
		switch(curEvent.type)
		{
			case SDL_KEYDOWN:
			{
				m_keyboardInput.PostKeyDownEvent(curEvent.key.keysym.sym);
			} break;

			case SDL_KEYUP:
			{
				m_keyboardInput.PostKeyUpEvent(curEvent.key.keysym.sym);
			} break;

			case SDL_VIDEORESIZE:
			{
//				g_graphicsInterface->ResizeScreen(curEvent.resize.w, curEvent.resize.h);
			} break;

			case SDL_QUIT:
			{
				FireQuitEvent();
			} break;
		}
	}

	m_mouseInput.PollMouse();
}

GraphicsContext& SDLWindow::GetGraphicsContext()
{
	return *m_graphicsContext;	
}

MouseInput& SDLWindow::GetMouseInput()
{
	return m_mouseInput;
}
