#include <OSWindow.h>
#include <GraphicsContext.h>
#include <MouseInput.h>
#include <KeyboardInput.h>

#include <System.h>
#include <R3Utils.h>

#include "Camera.h"

#include <Projection.h>

#include <GraphicsFrame.h>
#include <Model.h>
#include <OBJFile.h>
#include <Materials.h>

#include <ScopedFStream.h>
#include <Exceptions.h>

#include <math.h>
#include <iostream>
#include <fstream>


static bool s_forward = false;
static bool s_back = false;
static bool s_left = false;
static bool s_right = false;
static bool s_rotLeft = false;
static bool s_rotRight = false;

static bool s_toggleFullscreen = false;

static bool s_quitFlag = false;

static void ProcessKeyboardInput(const KeyboardInput::KeyEvent& key)
{
	switch(key.key)
	{
#ifdef USE_SDL
		case KeyboardInput::KT_W:
		{
			s_forward = key.pressed;
		} break;
		case KeyboardInput::KT_S:
		{
			s_back = key.pressed;
		} break;
		case KeyboardInput::KT_A:
		{
			s_left = key.pressed;
		} break;
		case KeyboardInput::KT_D:
		{
			s_right = key.pressed;
		} break;
#else
		case KeyboardInput::KT_COMMA:
		{
			s_forward = key.pressed;
		} break;
		case KeyboardInput::KT_O:
		{
			s_back = key.pressed;
		} break;
		case KeyboardInput::KT_A:
		{
			s_left = key.pressed;
		} break;
		case KeyboardInput::KT_E:
		{
			s_right = key.pressed;
		} break;
#endif
		case KeyboardInput::KT_UP:
		{
			s_rotLeft = key.pressed;
		} break;
		case KeyboardInput::KT_DOWN:
		{
			s_rotRight = key.pressed;
		} break;
		case KeyboardInput::KT_F:
		{
			s_toggleFullscreen = true;
		} break;

		case KeyboardInput::KT_ESC:
		{
			s_quitFlag = true;
		} break;
		
		default: break;
	};
}

static int s_mouseX = 0;
static int s_mouseY = 0;

static void ProcessMouseMove(const MouseInput::MoveEvent& moveEvent)
{
	s_mouseX += moveEvent.dx;
	s_mouseY += moveEvent.dy;
}


static void ProcessInput(float deltaT, Camera& cam)
{
	const float speed = 1.0f;

	if(s_forward)
	{
		Vector<4> forward(cam.GetRotMatrix().GetColumn(2));
		cam.SetPosition(cam.GetPosition() + forward * deltaT * speed);
	}
	if(s_back)
	{
		Vector<4> back(-cam.GetRotMatrix().GetColumn(2));
		cam.SetPosition(cam.GetPosition() + back * deltaT * speed);
	}
	if(s_left)
	{
		Vector<4> left(cam.GetRotMatrix().GetColumn(0));
		cam.SetPosition(cam.GetPosition() + left * deltaT * speed / 2);
	}
	if(s_right)
	{
		Vector<4> right(-cam.GetRotMatrix().GetColumn(0));
		cam.SetPosition(cam.GetPosition() + right * deltaT * speed / 2);
	}


	cam.AddPitch((float)s_mouseY / 1000.0f);
	cam.AddYaw((float)s_mouseX / 1000.0f);

	s_mouseY = 0;
	s_mouseX = 0;
}




static void MarkQuitFlag()
{
	s_quitFlag = true;
}

int main(int argc, char** args)
{
	OSWindow mainWin("GraphicsFrameTest", 1024, 768);

	mainWin.AttachQuitEventHandler(&MarkQuitFlag);

	KeyboardInput::KeyListenerConnection keyListener(mainWin.GetKeyboardInput(), &ProcessKeyboardInput);
	MouseInput::MoveListenerConnection mouseListener(mainWin.GetMouseInput(), &ProcessMouseMove);

	Camera myCam;
	Vector<4, VT_INSTANCE, false> lightPosVec(1, 3, 0.25, 0);

	myCam.SetPosition(Vector<4>(0,0,-4,1));
	
	try
	{
		MaterialSet materials;
		OBJFile testObj(ScopedFStream<std::ifstream>("Guy.obj").s, ".");
		Model testBox(testObj.GetObjectItem(0), AddMaterialsFromOBJFile(materials, testObj));

		float curSec = System::GetTimeInSeconds();
		while(!s_quitFlag)
		{
			if(s_toggleFullscreen)
			{
				s_toggleFullscreen = false;
				mainWin.SetWindowMode(mainWin.GetWindowMode() == OSWindow::M_WINDOWED ? OSWindow::M_FULLSCREEN : OSWindow::M_WINDOWED);
			}

			mainWin.GetGraphicsContext().FlushWait();

			GraphicsFrame frame(1000, 1000);

			float nextSec = System::GetTimeInSeconds();
			float deltaT = nextSec - curSec;
			curSec = nextSec;
			ProcessInput(deltaT, myCam);

			for(int i=0; i<5; ++i)
			{
				Matrix<4,4> modelWorld;
				modelWorld.GetVal(2,3) = i*5;

				testBox.AddToGraphicsFrame(frame, myCam.GetModelView()*modelWorld);
			}

			frame.Lock();
			mainWin.GetGraphicsContext().RenderGraphicsFrame(frame, MakePerspectiveProjectionMatrix(45.0f, mainWin.GetWidth()/mainWin.GetHeight(), 0.001f, 600.0f));
		}
	}
	catch(const ParseException& e)
	{
		std::cerr << "Parse error: " << e.GetMsg() << std::endl;
	}
	catch(ScopedFStream<std::ifstream>::IOException& e)
	{
		std::cerr << "IO error: " << e.GetMsg() << std::endl;
	}

	return 0;
}

