#ifndef TOOLWORLDDEFINITION_H_
#define TOOLWORLDDEFINITION_H_

#include "WorldDefinition.h"

#include <PotentialObject.h>
#include <ResourceManager.h>

#include <Model.h>

#include <SafePointerList.h>

#include <R3Utils.h>

#include <Events.h>

#include <map>
#include <list>


/**
 * Defines a rotation-scale-translation combination.
 * The CompositionTransformation is used to keep seperate the transformation data
 * that will eventually be combined in to a matrix (see the method ToMatrix).
 * This way we can rotate to our heart's content without brutalizing a matrix
 * through repeated matrix multiplication with near identity matrices.
 */
class CompositionTransformation
{
	UnalignedUnitQuaternion m_rot;
	Vector<3, VT_INSTANCE, false> m_pos;
	float m_scale;

public:
	CompositionTransformation()
		: m_pos(0,0,0)
		, m_scale(1.0f)
	{};
	CompositionTransformation(std::istream& in);
		

	void SetRotation(const UnitQuaternion& rot) {m_rot = rot;}
	const UnalignedUnitQuaternion& GetRotation() const {return m_rot;}

	void SetPosition(const Vector<3>& pos) {m_pos = pos;}
	const Vector<3, VT_INSTANCE, false>& GetPosition() const {return m_pos;}

	void SetScale(float scale) {m_scale = scale;}
	float GetScale() const {return m_scale;}

	Matrix<4,4> ToMatrix() const;

	void Save(std::ostream& out) const;
};

/**
 * ToolSymbol wraps the Symbol class to enable tool-specific data without modifying
 * the underlying WorldDefinition Symbol.
 */
class ToolSymbol
{
	std::string m_name;

	std::auto_ptr<Symbol> m_mySymbol; // Only valid if we weren't given a symbol to begin with, so that the ToolSymbol can be independent
	Symbol* m_sym;

	struct ModelInfo
	{
		ResourceIdentifier m_modelId;
		ResourceReference<Model, ModelFromObjFactory> m_modelRes;

		ModelInfo(ResourceManager& resMan, const ResourceIdentifier& resId)
			: m_modelId(resId)
			, m_modelRes(resMan, resId)
		{};
	};

	PotentialObject<ModelInfo> m_modelInfo;

	void RenderModelResource(GraphicsFrame* frame, const Matrix<4,4>* modelView, float alpha, int layer, bool wireframe);

public:
	ToolSymbol(const std::string& name)
		: m_name(name)
		, m_mySymbol(new Symbol())
		, m_sym(m_mySymbol.get())
	{}
		
	/**
	 * @param sym	The Symbol that this newly constructed ToolSymbol will be wrapping (and function call forwarding to).
	 */
	ToolSymbol(const std::string& name, std::auto_ptr<Symbol> sym)
		: m_name(name)
		, m_mySymbol(sym)
		, m_sym(m_mySymbol.get())
	{}

	std::auto_ptr<Symbol>& GetMySymbol() {return m_mySymbol;}
		
	Symbol* GetSymbol() {return m_sym;} /**< Returns the internal WorldDefinition Symbol that the ToolSymbol is wrapping */
	const Symbol* GetSymbol() const {return m_sym;}

	void SetModel(ResourceManager& resMan, const ResourceIdentifier& resId); /**< Sets the symbol's approximation to be the rendering of the passed in model */
	void ClearModel(); /**< Clears the symbol's set approximation method so it is approximated with nothing. */

	bool IsModelSet() const {return m_modelInfo.Exists();} /**< @return true if a Model has been set for this Symbol's approximation */

	/**
	 * @pre	IsModelSet() == true
	 * @return The ResourceIdentifier of the Model currently set as this Symbol's approximation.
	 */
	const ResourceIdentifier& GetModelResID() const {ASSERT(IsModelSet()); return m_modelInfo->m_modelId;}

	/** Gets the approximation Model.
	 * @pre IsModelSet() == true
	 * @return The ResourceReference we have to the model that we are currently using as this Symbol's approximation. 
	 */
	const ResourceReference<Model, ModelFromObjFactory>& GetModel() const {ASSERT(IsModelSet()); return m_modelInfo->m_modelRes;}

	/** Returns the name of this symbol.
	 * @return The name of this ToolSymbol (identifying it within a ToolWorldDefinition)
	 */
	const std::string& GetName() const {return m_name;}
	
};

/** 
 * ToolProduction encapsulates the Production object within this tool environment
 * In this way, we can maintain tool-specific data without modifying the underlying
 * WorldDefinition system.
 */
class ToolProduction
{
	std::auto_ptr<Production> m_myProduction;

	Production* m_prod;

	// This vector here parallels the Production::GetChildren() vector
	std::vector<CompositionTransformation> m_transList;

	ToolProduction(const ToolProduction& other);
public:
	ToolProduction(std::auto_ptr<Production> prod)
		: m_myProduction(prod)
		, m_prod(m_myProduction.get())
	{};

	/** Returns our production so that it can be entered in to the WorldDefinition system.
	 * When a ToolProduction is first created, it will own a Production object.  When
	 * the ToolProduction is entered in to a ToolWorldDefinition system though, it
	 * passes ownership of the internal Production to the WorldDefinition system.
	 * Despite this, we maintain a pointer to it through the m_prod variable.
	 * @return Ownership of the internal Production object.
	 */
	std::auto_ptr<Production> GetMyProduction() {return m_myProduction;}
		
	Production* GetProduction() {return m_prod;}
	const Production* GetProduction() const {return m_prod;}

	const CompositionTransformation& GetTransformation(int childIndex) const {return m_transList[childIndex];}

	void AddRS(const CompositionTransformation& trans, ToolSymbol* sym);
	void EditRS(int childIndex, const CompositionTransformation& trans);

	/** Removes an item from the right side of this production.
	 * @param index The index of the right hand side symbol we want to remove.
	 */
	void RemoveRS(int index);
};

class ResourceManager;

/**
 * ToolWorldDefinition is a wrapper around WorldDefinition that augments it with
 * extra information and functionality that the tool needs, such as associating 
 * names to symbols and productions, and the ability to save and load definitions.
 */
class ToolWorldDefinition
{
	ResourceManager& m_resMan;

	WorldDefinition m_worldDef;

	typedef std::map<std::string, ToolSymbol*> NameSymbolMap;
	typedef std::map<const ToolSymbol*, std::string> NameSymbolInverseMap;
	NameSymbolMap m_nameSymbolMap;
	NameSymbolInverseMap m_nameSymbolInverseMap;

	typedef std::map<const Production*, const ToolProduction*> ProdToToolProdMap;
	ProdToToolProdMap m_prodToToolProd;

	typedef std::map<const Symbol*, const ToolSymbol*> SymToToolSymMap;
	SymToToolSymMap m_symToToolSym;

	typedef SafePointerList<ToolSymbol> ToolSymbolList;
	ToolSymbolList m_symList;
	typedef SafePointerList<ToolProduction> ToolProductionList;
	ToolProductionList m_prodList;

	PotentialObject<boost::filesystem::path> m_workingPath;

	void AddSymbol(std::auto_ptr<ToolSymbol> toolSym);
	
	void Save(std::ostream& out) const;
	void SaveSymbol(std::ostream& out, const ToolSymbol& sym) const;
	void SaveProduction(std::ostream& out, const ToolProduction& prod) const;
	
	void Load(std::istream& in);
	void LoadSymbol(std::istream& in);
	void LoadProduction(std::istream& in);

public:
	/**
	 * @param sym A pointer to the ToolSymbol representing the Symbol that was just added.
	 */
	typedef void SymbolAddedSig(const ToolSymbol* sym);

	/**
	 * @param sym The symbol that was removed.  This pointer is effectively invalid at this time.
	 */
	typedef void SymbolRemovedSig(const ToolSymbol* sym);

	/**
	 * @param prod A pointer to the ToolProduction object that was just added.
	 */
	typedef void ProductionAddedSig(const ToolProduction* prod);

	/**
	 * @param prod A pointer to the production that was removed (although beware, the pointer is invalid at this time)
	 */
	typedef void ProductionRemovedSig(const ToolProduction* prod);

	/**
	 * @param prod A pointer to the ToolProduction object whos Right Side has become the proud owner of a new symbol.
	 * @param index The index in to the Production's array of right side Symbols that points to the newly added symbol.
	 */
	typedef void ProductionRSAddedSig(const ToolProduction* prod, int index);

	/**
	 * @param prod The production that had its Right Side pillaged.
	 * @param index The old index of the child that was removed.
	 */
	typedef void ProductionRSRemovedSig(const ToolProduction* prod, int index);

	/**
	 * @param prod A pointer to the ToolProduction object who has just changed.
	 */
	typedef void ProductionChangedSig(const ToolProduction* prod);

	/**
	 * @param sym The ToolSymbol that was modified.
	 */
	typedef void SymbolChangedSig(const ToolSymbol* sym);

private:
	Event<SymbolAddedSig> m_symbolAddedEvent;
	Event<SymbolRemovedSig> m_symbolRemovedEvent;
	Event<ProductionAddedSig> m_productionAddedEvent;
	Event<ProductionRemovedSig> m_productionRemovedEvent;
	Event<ProductionRSAddedSig> m_productionRSAddedEvent;
	Event<ProductionRSRemovedSig> m_productionRSRemovedEvent;
	Event<ProductionChangedSig> m_productionChangedEvent;
	Event<SymbolChangedSig> m_symbolChangedEvent;

public:
	/** Constructs an empty ToolWorldDefinition.
	 * @param resMan Reference to the ResourceManager that the ToolWorldDefinition will use
	 *					to manage any resources it requires (ie: models and materials)
	 */
	ToolWorldDefinition(ResourceManager& resMan);

	/** Constructs a ToolWorldDefinition by loading it from a file.
	 * @param resMan Reference to the ResourceManager that the ToolWorldDefinition will use
	 *					to manage any resources it requires (ie: models and materials)
	 * @param loadFrom The file to load the ToolWorldDefinition from.
	 */
	ToolWorldDefinition(ResourceManager& resMan, const boost::filesystem::path& loadFrom);
	~ToolWorldDefinition();

	/** Called when a Symbol is added to the WorldDefinition. */
	Event<SymbolAddedSig>::Port GetSymbolAddedEventPort()
	{
		return m_symbolAddedEvent.GetPort();
	}

	/** Called when a Symbol is removed from the WorldDefinition. */
	Event<SymbolRemovedSig>::Port GetSymbolRemovedEventPort()
	{
		return m_symbolRemovedEvent.GetPort();
	}

	/** Called when a Production is added to the WorldDefinition. */
	Event<ProductionAddedSig>::Port GetProductionAddedEventPort()
	{
		return m_productionAddedEvent.GetPort();
	}

	/** Called when a production has been removed. */
	Event<ProductionRemovedSig>::Port GetProductionRemovedEventPort()
	{
		return m_productionRemovedEvent.GetPort();
	}

	/** Called when a symbol has been added to the right side of a production. */
	Event<ProductionRSAddedSig>::Port GetProductionRSAddedEventPort()
	{
		return m_productionRSAddedEvent.GetPort();
	}

	/** Called when a symbol has had an element from its right side removed. */
	Event<ProductionRSRemovedSig>::Port GetProductionRSRemovedEventPort()
	{
		return m_productionRSRemovedEvent.GetPort();
	}

	/** Called when a property of a Production has changed (ie: it's scale/rotation/translation has changed). */
	Event<ProductionChangedSig>::Port GetProductionChangedEventPort()
	{
		return m_productionChangedEvent.GetPort();
	}

	/** Called when a property of a Symbol has changed (ie: It's approximation threshold has changed). */
	Event<SymbolChangedSig>::Port GetSymbolChangedEventPort()
	{
		return m_symbolChangedEvent.GetPort();
	}

	/** Returns the internally maintained WorldDefinition.
	 * @return The WorldDefinition object that this ToolWorldDefinition is wrapped around.
	 */
	const WorldDefinition& GetWorldDefinition() const {return m_worldDef;}
	
	/** Adds a symbol to the WorldDefinition.
	 *	This will add a Symbol to the internal WorldDefinition, as well as
	 *	wrap it with an internally created ToolSymbol object.
	 *	@param symName The name that will be associated with the newly added Symbol.
	 *	@param sym The Symbol that you wish to add to the WorldDefinition.
	 */
	void AddSymbol(const std::string& symName, std::auto_ptr<Symbol> sym);

	/** Removes the specified symbol from the WorldDefinition.
	 * @param sym The symbol to remove.  After the call, this pointer will be invalid.
	 */
	void RemoveSymbol(const ToolSymbol* sym);

	/** Retrieves a Symbol added to the WorldDefinition with the passed in name.
	 *	@param symName The name of the symbol you wish to retrieve.
	 *	@return A pointer to the found ToolSymbol, or 0 if none could be found matching the name.
	 */
	const ToolSymbol* GetSymbol(const std::string& symName);
	
	/** Retrieves a ToolSymbol that contains the original symbol.
	 *	@param sym The symbol that is wrapped by the ToolSymbol that will be returned.
	 *	@return A ToolSymbol that wraps the given symbol.
	 */
	const ToolSymbol* GetSymbol(const Symbol* sym);

	/** Retrieves a list of all Symbols in the WorldDefinition.
	 *	@return A list of the symbol names that have been added to this WorldDefinition.
	 */
	std::list<const ToolSymbol*> GetSymbolList() const;

	/** Given a Symbol, returns the string associated with it.
	 *	@param sym The Symbol that you want to find the name of
	 *	@return A string representing the ToolSymbol (which represents the Symbol you passed in)
	 */
	std::string GetStringForSymbol(const Symbol* sym) const;
	/** Given a ToolSymbol, returns the string associated with it.
	 *	@param sym The ToolSymbol that you want to find the name of
	 *	@return A string representing the ToolSymbol that you passed in.
	 */
	std::string GetStringForSymbol(const ToolSymbol* sym) const;

	/** Automatically produces a string describing a production.
	 *	Returns a string in the form "LS -> RS1 RS2 RS3 ..." of the symbols
	 *	that make up the passed in Production.
	 *	@param prod The Production you want to produce a name for.
	 *	@return A string describing the passed in Production
	 */
	std::string GetStringForProduction(const Production* prod) const;

	/** Given a Production, returns its representative ToolProduction.
	 *	@param prod The Production you want to lookup.
	 *	@return The ToolProduction corresponding to the passed in Production, in this ToolWorldDefinition environment.
	 */
	const ToolProduction* ToolProductionFromProduction(const Production* prod) const;

	/** Sets the approximation of a ToolSymbol object to be the model pointed to by the passed in ResourceIdentifier.
	 * When set, the ToolSymbol (and so, the corresponding Symbol) will be rendered with the model that is the resId resource.
	 * @param toolSym The ToolSymbol we wish to set the approximation model of.
	 * @param resId The ResourceIdentifier pointing to the Model resource we will use to approximate the Symbol with.
	 */
	void SetSymbolModel(const ToolSymbol* toolSym, const ResourceIdentifier& resId);

	/** Clears the set approximation model of the symbol so that it is approximated with nothing.
	 * @param toolSym the ToolSymbol we wish to clear the approximation of.
	 */
	void ClearSymbolModel(const ToolSymbol* toolSym);

	/** Sets the Symbol approximation threshold.
	 * @param	sym The Symbol to set the threshold of.
	 * @param	t	The threshold value to set for the wrapped Symbol.
	 */
	void SetSymbolApproxThreshold(const ToolSymbol* sym, float t);

	/** Adds a Production to the WorldDefinition.
	 *	@param sym The Symbol that will be used as the Left Side of the Production we're about to create
	 *	@param fadeMode The fade mode to be used by the Production to fade in and out of the approximation.
	 */
	const ToolProduction* AddProduction(const ToolSymbol* sym, Production::FadeMode fadeMode);

	/** Removes a production from the WorldDefinition.
	 * @param prod The production to remove.  The passed in pointer will be invalid after this call.
	 */
	void RemoveProduction(const ToolProduction* prod);

	/** Adds a Symbol to the Right Side of a Production.
	 * @param prod The Production who's Right Side we will add the Symbol, rsSym, to.
	 * @param rsSym The Symbol we're about to add to the Production's Right Side.
	 */
	void AddSymbolToProductionRS(const ToolProduction* prod, const ToolSymbol* rsSym);

	/** Removes a symbol from the Right Side of a production.
	 * @param prod The production whose right side we will remove a symbol from
	 * @param index The index of the right-hand side symbol that we should remove.
	 */
	void RemoveSymbolFromProductionRS(const ToolProduction* prod, int index);

	/** Edits the composition of a given Production.
	 *	This function allows you to edit the composition of a production, by changing the transform
	 *	to be applied to one of it's child symbols.
	 *	@param prod The Production that we wish to edit.
	 *	@param childNum The child index in to the Production's child list who's transformation we will edit.
	 *	@param transModifier This is a description of the transformation we wish to apply to the Production's child.
	 */
	void EditProductionComp(const ToolProduction* prod, int childNum, const CompositionTransformation& transModifier);

	/** Modifies the fade mode of the given production.
	 * @param prod The production you wish to edit.
	 * @param fadeMode The new fade mode to assign to the production.
	 */
	void EditProductionFadeMode(const ToolProduction* prod, Production::FadeMode fadeMode);
	
	/** Returns the working path that the ToolWorldDefinition is operating out of.
	 * The working path is where the ToolWorldDefinition file was saved to.  All resources
	 * used by the WorldDefinition will be relative to this path.
	 * @return The working path of the world definition.  It may not exist if this world definition has never been saved before.
	 */
	PotentialObject<boost::filesystem::path> GetWorkingPath() {return m_workingPath;}
	
	/** Saves the WorldDefinition to a file.
	 * When called, this function will serialize the ToolWorldDefinition to the given file.
	 * All resources that this WorldDefinition reference will have their path names saved
	 * relative to the file we save this WorldDefinition to (the saveTo parameter).
	 * After this function is called, GetWorkingPath() will return saveTo.
	 * @param saveTo The file you wish to save the WorldDefinition out to.
	 */
	void Save(const boost::filesystem::path& saveTo);
};

#endif
