#ifndef COMPOSITIONLIBRARY_H_
#define COMPOSITIONLIBRARY_H_

#include <Materials.h>
#include <Algebra.h>
#include <Exceptions.h>
#include <ResourceManager.h>
#include <Model.h>

#include <vector>
#include <list>
#include <memory>
#include <set>
#include <map>

#include <boost/function.hpp>

class GraphicsFrame;

class Model;

class Production;

/**
 * Represents a non-terminal or terminal symbol within the scope of a WorldDefinition.
 * A world is constructed out of Symbols and Productions.  The Symbols represent the
 * actual objects in the world, and can be approximated by rendering a model for them.
 * In the case that a symbol is approximated, it can be thought of as being a terminal
 * symbol (since no productions are applied to it to produce more symbols).  If the
 * symbol is not approximated, then it is expanded in to other symbols via a Production
 * created with this as a left hand side symbol.
 */
class Symbol
{
	public:
	typedef std::vector<Production*> ProductionArray;	
	typedef std::set<Production*> ProductionSet;

	private:
	ProductionSet m_lsProductions; // Productions this symbol is on the left of
	ProductionArray m_lsProductionsArray; // Those same symbols in an array
	ProductionSet m_rsProductions; // Productions this symbol is on the right of
	ProductionArray m_rsProductionsArray; // Same symbols in an array

	friend class WorldDefinition;
	friend class Production;

	public:
	typedef boost::function<void (GraphicsFrame* frame, const Matrix<4,4>* modelView, float alpha, int layer, bool wireframe)> ApproxRenderFunc;

	private:

	ApproxRenderFunc m_approxFunc;

	float m_areaApproximateThreshold;

	void AddLSProduction(Production* prod);
	void RemoveLSProduction(Production* prod);

	// TODO: These should allow multiple instances of the same production
	void AddRSProduction(Production* prod);
	void RemoveRSProduction(Production* prod);

	public:
	Symbol()
		: m_areaApproximateThreshold(0.02f)
	{};

	const ProductionSet& GetLSProductions() const {return m_lsProductions;} /**< Returns a set of productions that this symbol is on the left-side of. */
	const ProductionSet& GetRSProductions() const {return m_rsProductions;}

	/** Is this symbol on the right side of the given production?
	 * @param prod The production we want to know whether we're on the right side of or not
	 * @return true if we are on the right side of the production
	 */
	bool OnRSOf(const Production* prod) const {return m_rsProductions.find(const_cast<Production*>(prod)) != m_rsProductions.end(); }

	const ProductionArray& GetLSProductionsArray() const {return m_lsProductionsArray;} /**< Returns an array of productions that this symbol is on the left-side of. */
	const ProductionArray& GetRSProductionsArray() const {return m_rsProductionsArray;}

	float GetAreaApproximateThreshold() const {return m_areaApproximateThreshold;} /**< Returns the value (in units of screen-space area) at which this symbol will be approximated my rendering a model instead of recursed. */
	ApproxRenderFunc GetApproxRenderFunc() const {return m_approxFunc;} /**< Returns the function that is to be used to render this symbol, to approximate it. */

	void SetApproximationRenderFunc(ApproxRenderFunc func) {m_approxFunc = func;} /**< Defines how to render this symbol when it is decided that it should be approximated. */
	void SetAreaApproximateThreshold(float thresh) {m_areaApproximateThreshold = thresh;} /**< Sets the point at which to start approximating this symbol instead of applying a production to it.  When the symbol's maximum screen-space area is less than this value, it is approximated. TODO: What the fuck are the units for this? */
	
};

class Production
{
	public:
	enum FadeMode
	{
		FM_FADEIN,
		FM_FADEOUT,
		FM_NONE,
		FM_FADEINOUT,

		FM_NUM_FADEMODES,
	};

	private:
	Symbol* m_ls;
	FadeMode m_fadeMode;

	public:
	struct ChildDescriptor
	{
		Symbol* sym;
		Matrix<4,4> transform;
		Matrix<4,4> invTransform;

		Tuple<std::complex<double>, 3> eigenValues; // eigenvalues of the 3x3 part of the transform matrix
	};
	typedef std::vector<ChildDescriptor*> ChildList;
	private:

	ChildList m_children;
	
	friend class WorldDefinition;

	public:
	Production(Symbol* ls, FadeMode fadeMode)
		: m_ls(ls)
		, m_fadeMode(fadeMode)
	{};
	~Production()
	{
		for(ChildList::iterator iter = m_children.begin(); iter != m_children.end(); ++iter)
		{
			delete *iter;
		}
	}

	Symbol* GetLS() {return m_ls;}
	const Symbol* GetLS() const {return m_ls;}

	FadeMode GetFM() const {return m_fadeMode;}
	void SetFM(FadeMode fadeMode) {m_fadeMode = fadeMode;}
	
	void AddRS(const Matrix<4,4>& transform, Symbol* sym);
	void EditRS(int childIndex, const Matrix<4,4>& transform, Symbol* sym);

	/** Removes an item from the right side of this production.
	 * @param childIndex The index of the child we wish to remove.
	 */
	void RemoveRS(int childIndex);

	ChildList& GetChildren() {return m_children;}
	const ChildList& GetChildren() const {return m_children;}
	
};


class WorldDefinition
{
	public:
	typedef std::set<Symbol*> SymbolSet;
	typedef std::set<Production*> ProductionSet;
	
	private:
	SymbolSet m_symbols;	
	typedef std::vector<Symbol*> SymbolArray;
	SymbolArray m_symArray;

	ProductionSet m_productions;
	typedef std::vector<Production*> ProductionArray;
	ProductionArray m_prodArray; // So we can randomly choose one

	typedef std::map<std::string, Symbol*> SymbolMap;
	SymbolMap m_loadedSymbols; // Symbols that were added from a load

	ResourceManager& m_resMan;

	std::vector<ResourceReference<Model, ModelFromObjFactory> > m_modelRefs;

	public:
	WorldDefinition(ResourceManager& resMan);
	~WorldDefinition();

	class LoadException : public std::exception
	{
		std::string m_msg;
	public:
		LoadException(const std::string& msg)
			: m_msg(msg)
		{};
		virtual const char* what() const {return m_msg.c_str();}
	};

	// Load in a set of symbols and productions from a directory containing OBJ files and TXT files
	void LoadFromDirectory(const std::string& dirName);
	Symbol* GetLoadedSymbol(const std::string& symName) const
	{
		SymbolMap::const_iterator found = m_loadedSymbols.find(symName);
		if(found == m_loadedSymbols.end()) return 0;
		return found->second;
	}

	void AddSymbol(std::auto_ptr<Symbol> sym);

	/** Removes the symbol from the system.
	 * @pre The symbol is not on the left or right side of any productions.
	 * @param sym The symbol to remove.  After this call, the pointer will be invalid.
	 */
	void RemoveSymbol(Symbol* sym);

	void AddProduction(std::auto_ptr<Production> prod);

	/** Removes the production from the system.
	 * @param prod The production to remove.  After this call, the pointer will be invalid.
	 */
	void RemoveProduction(Production* prod);

	const SymbolSet& GetSymbols() const {return m_symbols;}
	const ProductionSet& GetProductions() const {return m_productions;}

	bool Contains(Symbol* sym) const;
	bool Contains(Production* prod) const;

	Symbol* GetRandomSymbol() const;
	Production* GetRandomProduction() const;
};

#endif
