#include "ToolWorldDefinition.h"
#include "PathRelativeTo.h"
#include <ResourceManager.h>

#include <R3Utils.h>

static const int s_version = 1;

ToolWorldDefinition::ToolWorldDefinition(ResourceManager& resMan)
: m_resMan(resMan)
, m_worldDef(m_resMan)
{
}

ToolWorldDefinition::~ToolWorldDefinition()
{
}

void ToolWorldDefinition::AddSymbol(std::auto_ptr<ToolSymbol> toolSym)
{
	NameSymbolMap::const_iterator found = m_nameSymbolMap.find(toolSym->GetName());
	ASSERT(found == m_nameSymbolMap.end());
	
	m_symToToolSym[toolSym->GetSymbol()] = toolSym.get();
	m_nameSymbolMap[toolSym->GetName()] = toolSym.get();
	m_nameSymbolInverseMap[toolSym.get()] = toolSym->GetName();
	
	m_worldDef.AddSymbol(toolSym->GetMySymbol());

	m_symList.objs.push_back(toolSym.release());
	
	m_symbolAddedEvent.Fire(m_symList.objs.back());
}


void ToolWorldDefinition::AddSymbol(const std::string& symName, std::auto_ptr<Symbol> sym)
{
	std::auto_ptr<ToolSymbol> toolSym(new ToolSymbol(symName, sym));

	AddSymbol(toolSym);
}

void ToolWorldDefinition::RemoveSymbol(const ToolSymbol* sym)
{
	SymToToolSymMap::iterator foundInSymToToolSym = m_symToToolSym.find(sym->GetSymbol());
	ASSERT(foundInSymToToolSym != m_symToToolSym.end());
	m_symToToolSym.erase(foundInSymToToolSym);

	NameSymbolMap::iterator foundInNameSymbolMap = m_nameSymbolMap.find(sym->GetName());
	ASSERT(foundInNameSymbolMap != m_nameSymbolMap.end());
	m_nameSymbolMap.erase(foundInNameSymbolMap);

	NameSymbolInverseMap::iterator foundInSymbolInverseMap = m_nameSymbolInverseMap.find(sym);
	ASSERT(foundInSymbolInverseMap != m_nameSymbolInverseMap.end());
	m_nameSymbolInverseMap.erase(foundInSymbolInverseMap);

	bool found = false;
	for(std::vector<ToolSymbol*>::iterator iter = m_symList.objs.begin(); iter != m_symList.objs.end(); ++iter)
	{
		if(*iter == sym)
		{
			found = true;
			m_symList.objs.erase(iter);
			break;
		}
	}
	ASSERT(found);

	m_symbolRemovedEvent.Fire(sym);

	m_worldDef.RemoveSymbol(const_cast<Symbol*>(sym->GetSymbol()));

	delete sym;
}

const ToolSymbol* ToolWorldDefinition::GetSymbol(const std::string& symName)
{
	NameSymbolMap::const_iterator found = m_nameSymbolMap.find(symName);

	if(found == m_nameSymbolMap.end()) return 0;

	return found->second;
}

const ToolSymbol* ToolWorldDefinition::GetSymbol(const Symbol* sym)
{
	SymToToolSymMap::const_iterator found = m_symToToolSym.find(sym);

	ASSERT(found != m_symToToolSym.end());

	return found->second;
}

std::string ToolWorldDefinition::GetStringForSymbol(const Symbol* sym) const
{
	SymToToolSymMap::const_iterator found = m_symToToolSym.find(sym);
	ASSERT(found != m_symToToolSym.end());

	return GetStringForSymbol(found->second);
}

std::string ToolWorldDefinition::GetStringForSymbol(const ToolSymbol* sym) const
{
	NameSymbolInverseMap::const_iterator found = m_nameSymbolInverseMap.find(sym);

	if(found == m_nameSymbolInverseMap.end()) return std::string("");

	return found->second;
}

const ToolProduction* ToolWorldDefinition::AddProduction(const ToolSymbol* sym, Production::FadeMode fadeMode)
{
	NameSymbolInverseMap::iterator found = m_nameSymbolInverseMap.find(sym);
	ASSERT(found != m_nameSymbolInverseMap.end());

	std::auto_ptr<Production> addedProd(new Production(const_cast<ToolSymbol*>(sym)->GetSymbol(), fadeMode));
	std::auto_ptr<ToolProduction> addedToolProd(new ToolProduction(addedProd));
	
	m_prodToToolProd[addedToolProd->GetProduction()] = addedToolProd.get();

	m_worldDef.AddProduction(addedToolProd->GetMyProduction());

	m_prodList.objs.push_back(addedToolProd.release());

	
	m_productionAddedEvent.Fire(m_prodList.objs.back());

	return m_prodList.objs.back();
}

void ToolWorldDefinition::RemoveProduction(const ToolProduction* prod)
{
	ToolProduction* toolProd = const_cast<ToolProduction*>(prod);
	Production* baseProd = toolProd->GetProduction();

	// Remove it from the prod to tool prod map
	ProdToToolProdMap::iterator prodToToolProdFound = m_prodToToolProd.find(baseProd);
	ASSERT(prodToToolProdFound != m_prodToToolProd.end());
	m_prodToToolProd.erase(prodToToolProdFound);

	// Remove it from the list of tool prods
	bool foundInProdList = false;
	for(std::vector<ToolProduction*>::iterator iter = m_prodList.objs.begin(); iter != m_prodList.objs.end(); ++iter)
	{
		if(toolProd == *iter)
		{
			foundInProdList = true;
			m_prodList.objs.erase(iter);
			break;
		}
	}
	ASSERT(foundInProdList);

	// Event firing time!
	m_productionRemovedEvent.Fire(toolProd);

	// We have no more use for the ToolProduction, delete it
	delete toolProd;

	// And finally remove the production from the WorldDefinition
	m_worldDef.RemoveProduction(baseProd);
}

std::string ToolWorldDefinition::GetStringForProduction(const Production* prod) const
{
	std::string prodString;

	{
		// Add the left hand side symbol to the string

		SymToToolSymMap::const_iterator found = m_symToToolSym.find(prod->GetLS());
		ASSERT(found != m_symToToolSym.end());

		prodString = GetStringForSymbol(found->second);
	}

	prodString += " ->";

	const Production::ChildList& children = prod->GetChildren();
	for(Production::ChildList::const_iterator iter = children.begin(); iter != children.end(); ++iter)
	{
		SymToToolSymMap::const_iterator found = m_symToToolSym.find((*iter)->sym);
		ASSERT(found != m_symToToolSym.end());

		prodString += " ";
		prodString += GetStringForSymbol(found->second);
	}

	return prodString;
}

void ToolProduction::AddRS(const CompositionTransformation& trans, ToolSymbol* sym)
{
	m_transList.push_back(trans);
	m_prod->AddRS(trans.ToMatrix(), sym->GetSymbol());
}

void ToolProduction::EditRS(int childIndex, const CompositionTransformation& trans)
{
	m_transList[childIndex] = trans;
	m_prod->EditRS(childIndex, trans.ToMatrix(), m_prod->GetChildren()[childIndex]->sym);
}


void ToolWorldDefinition::AddSymbolToProductionRS(const ToolProduction* prod, const ToolSymbol* rsSym)
{
	ToolProduction* workProd = const_cast<ToolProduction*>(prod);

	CompositionTransformation defaultTrans;
	defaultTrans.SetPosition(Vector<3>(0.5f, 0.5f, 0.5f));
	defaultTrans.SetScale(0.5f);

	workProd->AddRS(defaultTrans, const_cast<ToolSymbol*>(rsSym));

	m_productionRSAddedEvent.Fire(prod, workProd->GetProduction()->GetChildren().size()-1);
}

std::list<const ToolSymbol*> ToolWorldDefinition::GetSymbolList() const
{
	std::list<const ToolSymbol*> symList;

	for(NameSymbolMap::const_iterator iter = m_nameSymbolMap.begin(); iter != m_nameSymbolMap.end(); ++iter)
	{
		symList.push_back(iter->second);
	}

	return symList;
}

void ToolWorldDefinition::EditProductionComp(const ToolProduction* prod, int childNum, const CompositionTransformation& trans)
{
	ToolProduction* workProd = const_cast<ToolProduction*>(prod);

	workProd->EditRS(childNum, trans);

	m_productionChangedEvent.Fire(prod);
}

void ToolWorldDefinition::EditProductionFadeMode(const ToolProduction* prod, Production::FadeMode fadeMode)
{
	ToolProduction* workProd = const_cast<ToolProduction*>(prod);
	
	workProd->GetProduction()->SetFM(fadeMode);
	
	m_productionChangedEvent.Fire(prod);
}


const ToolProduction* ToolWorldDefinition::ToolProductionFromProduction(const Production* prod) const
{
	ProdToToolProdMap::const_iterator found = m_prodToToolProd.find(prod);
	ASSERT(found != m_prodToToolProd.end());

	return found->second;
}

Matrix<4,4> CompositionTransformation::ToMatrix() const
{
	Matrix<4,4> ret(Matrix<4,4>::Identity);

	float invScale = 1 / m_scale;

	ret = MakeTranslationMatrix(m_pos)
		* m_rot.ToMatrix() 		
		* MakeTranslationMatrix(Vector<3>(-0.5,-0.5,-0.5)*m_scale)
		* MakeUniScaleMatrix(m_scale);
	
	return ret;
}

void ToolWorldDefinition::SetSymbolModel(const ToolSymbol* toolSym, const ResourceIdentifier& resId)
{
	const_cast<ToolSymbol*>(toolSym)->SetModel(m_resMan, resId);

	m_symbolChangedEvent.Fire(toolSym);
}

void ToolWorldDefinition::ClearSymbolModel(const ToolSymbol* toolSym)
{
	const_cast<ToolSymbol*>(toolSym)->ClearModel();

	m_symbolChangedEvent.Fire(toolSym);
}


void ToolWorldDefinition::SetSymbolApproxThreshold(const ToolSymbol* sym, float t)
{
	const_cast<ToolSymbol*>(sym)->GetSymbol()->SetAreaApproximateThreshold(t);

	m_symbolChangedEvent.Fire(sym);
}

void ToolSymbol::SetModel(ResourceManager& resMan, const ResourceIdentifier& resId)
{
	if(m_modelInfo.Exists()) m_modelInfo.Clear();

	new(m_modelInfo) ModelInfo(resMan, resId);

	m_sym->SetApproximationRenderFunc(boost::bind(&ToolSymbol::RenderModelResource, this, _1, _2, _3, _4, _5));
}

void ToolSymbol::ClearModel()
{
	m_sym->SetApproximationRenderFunc(Symbol::ApproxRenderFunc());

	if(m_modelInfo.Exists()) m_modelInfo.Clear();
}

void ToolSymbol::RenderModelResource(GraphicsFrame* frame, const Matrix<4,4>* modelView, float alpha, int layer, bool wireframe)
{
	ASSERT(m_modelInfo.Exists());

	m_modelInfo->m_modelRes->AddToGraphicsFrame(*frame, *modelView, alpha, layer, wireframe);
}

void ToolWorldDefinition::Save(const boost::filesystem::path& saveTo)
{
	if(m_workingPath.Exists()) m_workingPath.Clear();
	
	new(m_workingPath) boost::filesystem::path(saveTo);
	ASSERT(m_workingPath->has_root_path());

	std::ofstream ofs(saveTo.string().c_str(), std::ios::binary);
	if(!ofs.is_open())
	{
		return;
	}

	Save(ofs);
}

ToolWorldDefinition::ToolWorldDefinition(ResourceManager& resMan, const boost::filesystem::path& loadFrom)
	: m_resMan(resMan)
	, m_worldDef(m_resMan)
{
	new(m_workingPath) boost::filesystem::path(loadFrom);
	
	// If it doesn't already have a root path, give it one by prepending the working directory
	if(!m_workingPath->has_root_path())
	{
		*m_workingPath = boost::filesystem::current_path() / *m_workingPath;
	}
	ASSERT(m_workingPath->has_root_path());

	std::ifstream ifs(loadFrom.string().c_str(), std::ios::binary);
	if(!ifs.is_open())
	{
		ASSERT(false);
	}
	
	Load(ifs);
}

template <class T>
static void OutData(std::ostream& out, const T& t)
{
	out.write(reinterpret_cast<const char*>(&t), sizeof(t));
}

template <class T>
static void InData(std::istream& in, T& t)
{
	in.read(reinterpret_cast<char*>(&t), sizeof(t));
}


static void OutString(std::ostream& out, const std::string& str)
{
	int strLen = str.size();
	OutData(out, strLen);
	out.write(str.c_str(), strLen);
}

static std::string InString(std::istream& in)
{
	int strLen;
	InData(in, strLen);
	
	char* strBuf = new char[strLen+1];
	in.read(strBuf, strLen);
	strBuf[strLen] = '\0';
	
	std::string retString(strBuf);
	
	delete[] strBuf;
	
	return retString;
}

CompositionTransformation::CompositionTransformation(std::istream& in)
{
	InData(in, m_rot);
	InData(in, m_pos);
	InData(in, m_scale);
}

void CompositionTransformation::Save(std::ostream& out) const
{
	OutData(out, m_rot);
	OutData(out, m_pos);
	OutData(out, m_scale);
}


void ToolWorldDefinition::Save(std::ostream& out) const
{
	OutData(out, s_version);

	int numSyms = m_symList.objs.size();
	OutData(out, numSyms);
	for(std::vector<ToolSymbol*>::const_iterator iter = m_symList.objs.begin(); iter != m_symList.objs.end(); ++iter)
	{
		SaveSymbol(out, **iter);
	}
	
	int numProds = m_prodList.objs.size();
	OutData(out, numProds);
	for(std::vector<ToolProduction*>::const_iterator iter = m_prodList.objs.begin(); iter != m_prodList.objs.end(); ++iter)
	{
		SaveProduction(out, **iter);
	}
}

void ToolWorldDefinition::SaveSymbol(std::ostream& out, const ToolSymbol& sym) const
{
	OutString(out, sym.GetName());
	
	OutData(out, sym.IsModelSet());
	if(sym.IsModelSet())
	{
		OutString(out, PathRelativeTo(*m_workingPath, sym.GetModelResID().GetFilepath()).string());
	}
	
	OutData(out, sym.GetSymbol()->GetAreaApproximateThreshold());
}

void ToolWorldDefinition::SaveProduction(std::ostream& out, const ToolProduction& prod) const
{
	OutData(out, prod.GetProduction()->GetFM());
	
	// Write the name of the Left Side symbol for this production
	OutString(out, GetStringForSymbol(prod.GetProduction()->GetLS()));
	
	int numChildren = prod.GetProduction()->GetChildren().size();
	OutData(out, numChildren);

	const Production::ChildList& children = prod.GetProduction()->GetChildren();
	
	int childIndex = 0;
	for(Production::ChildList::const_iterator iter = children.begin(); iter != children.end(); ++iter, ++childIndex)
	{
		OutString(out, GetStringForSymbol((*iter)->sym));

		prod.GetTransformation(childIndex).Save(out);
	}
}

void ToolWorldDefinition::Load(std::istream& in)
{
	int version;
	InData(in, version);

	int numSyms;
	InData(in, numSyms);
	for(int i=0; i<numSyms; ++i)
	{
		LoadSymbol(in);
	}
	
	int numProds;
	InData(in, numProds);
	for(int i=0; i<numProds; ++i)
	{
		LoadProduction(in);
	}
}

void ToolWorldDefinition::LoadSymbol(std::istream& in)
{
	std::string symName = InString(in);
	
	std::auto_ptr<ToolSymbol> loadSym(new ToolSymbol(symName));

	bool isModelSet;
	InData(in, isModelSet);
	if(isModelSet)
	{
		std::string modelFile = InString(in);
		ASSERT(!boost::filesystem::path(modelFile).has_root_path());

		loadSym->SetModel(m_resMan, ResourceIdentifier(m_workingPath->branch_path() / modelFile));
	}
	
	float approxThresh;
	InData(in, approxThresh);
	
	loadSym->GetSymbol()->SetAreaApproximateThreshold(approxThresh);

	AddSymbol(loadSym);
}

void ToolWorldDefinition::LoadProduction(std::istream& in)
{
	Production::FadeMode fm;
	InData(in, fm);
	
	std::string lsName = InString(in);
	const ToolSymbol* lsSym = GetSymbol(lsName);
	
	ToolProduction* toolProd = const_cast<ToolProduction*>(AddProduction(lsSym, fm));

	int numChildren;
	InData(in, numChildren);
	
	for(int i=0; i<numChildren; ++i)
	{
		// Add the children
		std::string childName = InString(in);
		const ToolSymbol* rsSym = GetSymbol(childName);
		
		AddSymbolToProductionRS(toolProd, rsSym);

		CompositionTransformation trans(in);
		toolProd->EditRS(i, trans);
	}
}

void ToolWorldDefinition::RemoveSymbolFromProductionRS(const ToolProduction* prod, int index)
{
	const_cast<ToolProduction*>(prod)->RemoveRS(index);

	m_productionRSRemovedEvent.Fire(prod, index);
}

void ToolProduction::RemoveRS(int index)
{
	std::vector<CompositionTransformation>::iterator iter;
	int i=0;
	for(iter = m_transList.begin(); iter != m_transList.end(); ++iter, ++i)
	{
		if(i == index) break;
	}
	ASSERT(iter != m_transList.end());

	m_transList.erase(iter);

	m_prod->RemoveRS(index);
}
