#include "MainSystemFrame.h"
#include "WorldDefinition.h"

#include "wxWidgetMouseInput.h"
#include "wxWidgetKeyboardInput.h"
#include "wxWidgetGraphicsContext.h"

#include "SymbolViewPanel.h"
#include <ResourceManager.h>

#include <SceneExplorer.h>

#include <OSWindow.h>

#include <wx/glcanvas.h>

#include <boost/bind.hpp>

BEGIN_EVENT_TABLE(MainSystemFrame, wxFrame)
	EVT_BUTTON(ADD_SYMBOL_BUTTON, MainSystemFrame::OnAddSymbol)
	EVT_BUTTON(REM_SYMBOL_BUTTON, MainSystemFrame::OnRemoveSymbol)
	EVT_LISTBOX(SYMBOL_LISTBOX, MainSystemFrame::OnSymbolSelected)
	EVT_MENU(ID_NEW, MainSystemFrame::OnNew)
	EVT_MENU(ID_SAVEAS, MainSystemFrame::OnSaveAs)
	EVT_MENU(ID_SAVE, MainSystemFrame::OnSave)
	EVT_MENU(ID_OPEN, MainSystemFrame::OnOpen)
	EVT_MENU(ID_QUIT, MainSystemFrame::OnQuit)
	EVT_MENU(ID_PREVIEW, MainSystemFrame::OnPreview)
END_EVENT_TABLE()

#include "wxWidgetKeyboardInput.h"

MainSystemFrame::MainSystemFrame(const wxString& title, const wxPoint& pos, const wxSize& size, ResourceManager& resMan)
: wxFrame((wxFrame*)NULL, -1, title, pos, size)
, m_symbolView(0)
, m_resMan(resMan)
{
	// Add the menu
	wxMenu* menuFile = new wxMenu();

	menuFile->Append( ID_NEW, "&New");
	menuFile->Append( ID_SAVEAS, "Save As...");
	menuFile->Append( ID_SAVE, "&Save");
	menuFile->Append( ID_OPEN, "&Open");
	menuFile->Append( ID_QUIT, "&Quit");
		
	wxMenu* menuView = new wxMenu();
	menuView->Append( ID_PREVIEW, "&Preview");

	wxMenuBar* menuBar = new wxMenuBar();
	menuBar->Append(menuFile, "&File");
	menuBar->Append(menuView, "&View");

	SetMenuBar(menuBar);
	


	wxPanel* rootPanel = new wxPanel(this, wxID_ANY);

	wxBoxSizer* bigHorizSizer = new wxBoxSizer(wxHORIZONTAL); // The sizer that splits the symbol list and the selected symbol panel
	rootPanel->SetSizer(bigHorizSizer);

	wxPanel* worldDefPanel = new wxPanel(rootPanel, wxID_ANY);
	bigHorizSizer->Add(worldDefPanel, 0, wxEXPAND|wxALL, 5);

	wxBoxSizer* vertSizer = new wxBoxSizer(wxVERTICAL);
	worldDefPanel->SetSizer(vertSizer);

	vertSizer->Add(new wxStaticText(worldDefPanel, -1, wxString("Symbols:")), 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

	m_symbolList = new wxListBox(worldDefPanel, SYMBOL_LISTBOX);
	vertSizer->Add(m_symbolList, 1, wxEXPAND|wxALL, 5);

	wxPanel* bottomButtonsPanel = new wxPanel(worldDefPanel, -1);
	vertSizer->Add(bottomButtonsPanel, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

	wxBoxSizer* bottomButtonsSizer = new wxBoxSizer(wxHORIZONTAL);
	bottomButtonsPanel->SetSizer(bottomButtonsSizer);

	wxButton* addSymbolButton = new wxButton(bottomButtonsPanel, ADD_SYMBOL_BUTTON, wxString("Add Symbol"));
	bottomButtonsSizer->Add(addSymbolButton, 1, wxALIGN_CENTER_VERTICAL|wxALL, 5);

	wxButton* remSymbolButton = new wxButton(bottomButtonsPanel, REM_SYMBOL_BUTTON, wxString("Remove Symbol"));
	bottomButtonsSizer->Add(remSymbolButton, 1, wxALIGN_CENTER_VERTICAL|wxALL, 5);

	m_selectedSymbolPlaceholder = new wxPanel(rootPanel, wxID_ANY);
	m_symbolPlaceholderSizer = new wxBoxSizer(wxVERTICAL);
	m_selectedSymbolPlaceholder->SetSizer(m_symbolPlaceholderSizer);

	bigHorizSizer->Add(m_selectedSymbolPlaceholder, 1, wxEXPAND|wxALL, 5);

	CreateWorldDefinition();
}

MainSystemFrame::~MainSystemFrame()
{
	if(m_symbolView != 0)
	{
		delete m_symbolView;
		m_symbolView = 0;
	}

	DestroyWorldDefinition();
}

void MainSystemFrame::OnWorldDefinitionCreated()
{
	// Add the existing symbols in the world definiton to our repertoire
	std::list<const ToolSymbol*> toolList = m_world->GetSymbolList();
	for(std::list<const ToolSymbol*>::const_iterator iter = toolList.begin(); iter != toolList.end(); ++iter)
	{
		HandleSymbolAdded(*iter);
	}

	new(m_symAddedConn) Event<ToolWorldDefinition::SymbolAddedSig>::Connection(m_world->GetSymbolAddedEventPort(), boost::bind(&MainSystemFrame::HandleSymbolAdded, this, _1));
	new(m_symRemovedConn) Event<ToolWorldDefinition::SymbolRemovedSig>::Connection(m_world->GetSymbolRemovedEventPort(), boost::bind(&MainSystemFrame::HandleSymbolRemoved, this, _1));
}

void MainSystemFrame::CreateWorldDefinition()
{
	new(m_world) ToolWorldDefinition(m_resMan);

	OnWorldDefinitionCreated();

	UpdateWindowCaption();
}
void MainSystemFrame::LoadWorldDefinition(const boost::filesystem::path& loadFrom)
{
	new(m_world) ToolWorldDefinition(m_resMan, loadFrom);

	OnWorldDefinitionCreated();

	UpdateWindowCaption();
}

void MainSystemFrame::DestroyWorldDefinition()
{
	m_symAddedConn.Clear();
	m_symRemovedConn.Clear();

	m_symbolList->Clear();

	m_world.Clear();
}

void MainSystemFrame::UpdateWindowCaption()
{
	if(m_world.Exists() && m_world->GetWorkingPath().Exists())
	{
		SetTitle(wxString("SceneComposer - " + m_world->GetWorkingPath()->string()));
	}
	else
	{
		SetTitle(wxString("SceneComposer - [untitled]"));
	}
}


void MainSystemFrame::HandleSymbolAdded(const ToolSymbol* sym)
{
	m_symbolList->Append(sym->GetName().c_str());
}

void MainSystemFrame::HandleSymbolRemoved(const ToolSymbol* sym)
{
	// Search for this symbol in our symbol list and pull it out
	for(int i=0; i<m_symbolList->GetCount(); ++i)
	{
		if(m_symbolList->GetString(i) == sym->GetName().c_str())
		{
			if(m_symbolList->GetSelection() == i)
			{
				if(m_symbolView != 0) DestroySymbolViewPanel();
			}

			m_symbolList->Delete(i);
			break;
		}
	}
}

AddSymbolDialog::AddSymbolDialog(wxWindow* parent, wxWindowID id)
	: wxDialog(parent, id, wxString("Add Symbol"),wxDefaultPosition, wxSize(400, 120))
{
	wxBoxSizer* vertSizer = new wxBoxSizer(wxVERTICAL);
	SetSizer(vertSizer);

	wxPanel* symNamePanel = new wxPanel(this, -1);
	wxBoxSizer* symNameSizer = new wxBoxSizer(wxHORIZONTAL);
	symNamePanel->SetSizer(symNameSizer);
	vertSizer->Add(symNamePanel, 1, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

	symNameSizer->Add(new wxStaticText(symNamePanel, -1, wxString("Symbol name:")), 0, wxALIGN_CENTER_VERTICAL|wxALL);

	m_symNameBox = new wxTextCtrl(symNamePanel, -1, wxString(""), wxDefaultPosition, wxSize(150, -1));
	symNameSizer->Add(m_symNameBox, 1, wxALIGN_CENTER_VERTICAL|wxALL, 5);

	
	wxPanel* buttonPanel = new wxPanel(this, -1);
	wxBoxSizer* buttonSizer = new wxBoxSizer(wxHORIZONTAL);
	buttonPanel->SetSizer(buttonSizer);
	vertSizer->Add(buttonPanel, 1, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

	wxButton* okButton = new wxButton(buttonPanel, wxID_OK, wxString("Okay"));
	buttonSizer->Add(okButton, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

	wxButton* cancelButton = new wxButton(buttonPanel, wxID_CANCEL, wxString("Cancel"));
	buttonSizer->Add(cancelButton, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

}

wxString AddSymbolDialog::GetSymbolName() const
{
	return m_symNameBox->GetValue();
}

void MainSystemFrame::OnAddSymbol(wxCommandEvent& WXUNUSED(event))
{
	AddSymbolDialog addSym(this, -1);

	if(addSym.ShowModal() == wxID_OK)
	{
		// Add the symbol to our list of symbols
		wxString chosenSymName = addSym.GetSymbolName();

		if(m_world->GetSymbol(chosenSymName.c_str()) != 0)
		{
			wxMessageBox("A symbol by that name already exists", "Error adding symbol", wxOK, this);
		}
		else
		{

			Symbol* newSym = new Symbol();
			m_world->AddSymbol(chosenSymName.c_str(), std::auto_ptr<Symbol>(newSym));
		}
	}
}

void MainSystemFrame::OnRemoveSymbol(wxCommandEvent& WXUNUSED(event))
{
	if(m_symbolList->GetSelection() == wxNOT_FOUND) return;

	const ToolSymbol* sym = m_world->GetSymbol(m_symbolList->GetStringSelection().c_str());

	if(sym->GetSymbol()->GetLSProductions().size() > 0
		|| sym->GetSymbol()->GetRSProductions().size() > 0)
	{
		wxMessageBox("Symbols cannot be deleted if they are on the left or right side of a production.  Delete them there first.", "Error removing symbol", wxOK, this);
		return;
	}

	m_world->RemoveSymbol(sym);
}

void MainSystemFrame::CreateSymbolViewPanel(const char* symbolName)
{
	m_symbolView = new SymbolViewPanel(*m_world, symbolName, m_selectedSymbolPlaceholder, this);
	m_symbolPlaceholderSizer->Add(m_symbolView, 1, wxEXPAND, 0);

	m_symbolPlaceholderSizer->Layout();
}

void MainSystemFrame::DestroySymbolViewPanel()
{
	ASSERT(m_symbolView != 0);
	
	m_symbolPlaceholderSizer->Remove(m_symbolView);
	delete m_symbolView;
	m_symbolView = 0;
}

	
void MainSystemFrame::OnSymbolSelected(wxCommandEvent& WXUNUSED(event))
{
	if(m_symbolView != 0)
	{
		if(m_symbolList->GetStringSelection() == m_symbolView->GetSymbolName()) return; // Nothing to do if we pick the same symbol again
	
		DestroySymbolViewPanel();
	}

	if(m_symbolList->GetStringSelection() != "")
	{
		CreateSymbolViewPanel(m_symbolList->GetStringSelection().c_str());
	}
}

void MainSystemFrame::OnNew(wxCommandEvent& WXUNUSED(event))
{
	if(m_symbolView != 0) DestroySymbolViewPanel();
	
	DestroyWorldDefinition();
	CreateWorldDefinition();
}

void MainSystemFrame::OnSaveAs(wxCommandEvent& WXUNUSED(event))
{
	wxFileDialog saveDialog(this, wxString("Save As"), wxEmptyString, wxEmptyString, wxString("Linfinity World Files (*.lin)|*.lin"), wxFD_SAVE);
	if(saveDialog.ShowModal() == wxID_OK)
	{
		// Do the save!
		m_world->Save(saveDialog.GetPath().c_str());

		UpdateWindowCaption();
	}
}

void MainSystemFrame::OnSave(wxCommandEvent& event)
{
	if(m_world->GetWorkingPath().Exists())
	{
		m_world->Save(*m_world->GetWorkingPath());
	}
	else
	{
		// We have never saved this before, go through the Save As flow
		OnSaveAs(event);
	}
}

void MainSystemFrame::OnOpen(wxCommandEvent& WXUNUSED(event))
{
	wxFileDialog openDialog(this, wxString("Open"), wxEmptyString, wxEmptyString, wxString("Linfinity World Files (*.lin)|*.lin"), wxFD_OPEN);
	
	if(openDialog.ShowModal() == wxID_OK)
	{
		if(m_symbolView != 0) DestroySymbolViewPanel();

		DestroyWorldDefinition();
		LoadWorldDefinition(openDialog.GetPath().c_str());
	}
}

void MainSystemFrame::OnQuit(wxCommandEvent& WXUNUSED(event))
{
	Close( TRUE );
}

/** Decorates a MouseInput object with the ability to pause the input event stream. */
class PauseableMouseInput : public MouseInput
{
	MouseInput& m_decoratee;

	PotentialObject<MouseInput::MoveListenerConnection> m_moveConn;
	PotentialObject<MouseInput::ButtonListenerConnection> m_buttonConn;

	bool m_paused;

	void OnMove(const MouseInput::MoveEvent& evt)
	{
		if(!m_paused) PostMoveEvent(evt);
	}
	void OnButton(const MouseInput::ButtonEvent& evt)
	{
		if(!m_paused) PostButtonEvent(evt);
	}
public:
	PauseableMouseInput(MouseInput& decoratee)
		: m_decoratee(decoratee)
		, m_paused(false)
	{
		new(m_moveConn) MouseInput::MoveListenerConnection(m_decoratee, boost::bind(&PauseableMouseInput::OnMove, this, _1));
		new(m_buttonConn) MouseInput::ButtonListenerConnection(m_decoratee, boost::bind(&PauseableMouseInput::OnButton, this, _1));
	}

	void SetPaused(bool paused) {m_paused = paused;}
	bool GetPaused() const {return m_paused;}
};

class PreviewWindow : public wxDialog
{
	class PreviewPanel : public wxGLCanvas
	{
		wxWindow& m_parent;

		PotentialObject<wxWidgetGraphicsContext> m_graphicsContext;
		
		PotentialObject<wxWidgetKeyboardInput> m_keyboardInput;

		PotentialObject<wxWidgetMouseInput> m_mouseInput;
		PotentialObject<PauseableMouseInput> m_pauseableMouseInput;

		void OnResize(wxSizeEvent& event)
		{
			event.Skip();

			int width;
			int height;
			GetClientSize(&width, &height);

			m_graphicsContext->ResizeScreen(width, height);
		}

		void OnLeftDown(wxMouseEvent& event)
		{
			if(!HasCapture())
			{
				CaptureMouse();
			}

			m_pauseableMouseInput->SetPaused(false);

			event.Skip();
		}
		void OnLeftUp(wxMouseEvent& event)
		{
			if(HasCapture())
			{
				ReleaseMouse();
			}

			m_pauseableMouseInput->SetPaused(true);

			event.Skip();
		}

		void OnMouseCaptureLost(wxMouseCaptureLostEvent& event)
		{
			m_pauseableMouseInput->SetPaused(true);
		}

		void OnKeyDown(wxKeyEvent& event)
		{
			if(event.GetKeyCode() == WXK_ESCAPE)
			{
				m_parent.Close();
			}

			event.Skip();
		}

	public:
		PreviewPanel(wxWindow* parent)
			: wxGLCanvas(parent, -1, 0, wxDefaultPosition, wxDefaultSize, wxCLIP_CHILDREN|wxFULL_REPAINT_ON_RESIZE)
			, m_parent(*parent)
		{
			new(m_graphicsContext) wxWidgetGraphicsContext(*this);
			
			new(m_keyboardInput) wxWidgetKeyboardInput(*this);

			new(m_mouseInput) wxWidgetMouseInput(*this);
			new(m_pauseableMouseInput) PauseableMouseInput(*m_mouseInput);
			m_pauseableMouseInput->SetPaused(true);
		};

		GraphicsContext& GetGraphicsContext() {return *m_graphicsContext;}
		MouseInput& GetMouseInput() {return *m_pauseableMouseInput;}
		KeyboardInput& GetKeyboardInput() {return *m_keyboardInput;}

		DECLARE_EVENT_TABLE()
	};
	PreviewPanel* m_panel;

	PotentialObject<SceneExplorer> m_sceneExplorer;
	const ToolWorldDefinition& m_world;

public:
	PreviewWindow(wxWindow* parent, const ToolWorldDefinition& world)
		: wxDialog(parent, -1, wxString("Preview"),wxDefaultPosition, wxSize(800, 600), wxDEFAULT_DIALOG_STYLE|wxRESIZE_BORDER|wxMAXIMIZE_BOX)
		, m_world(world)
	{

		wxBoxSizer* sizer = new wxBoxSizer(wxHORIZONTAL);
		SetSizer(sizer);

		m_panel = new PreviewPanel(this);
		sizer->Add(m_panel, 1, wxEXPAND);

		new(m_sceneExplorer) SceneExplorer(m_world.GetWorldDefinition(), m_panel->GetGraphicsContext(), m_panel->GetKeyboardInput(), m_panel->GetMouseInput());
	}
};
BEGIN_EVENT_TABLE(PreviewWindow::PreviewPanel, wxGLCanvas)
	EVT_SIZE(PreviewWindow::PreviewPanel::OnResize)
	EVT_LEFT_DOWN(PreviewWindow::PreviewPanel::OnLeftDown)
	EVT_LEFT_UP(PreviewWindow::PreviewPanel::OnLeftUp)
	EVT_KEY_DOWN(PreviewWindow::PreviewPanel::OnKeyDown)
	EVT_MOUSE_CAPTURE_LOST(PreviewWindow::PreviewPanel::OnMouseCaptureLost)
END_EVENT_TABLE()


void MainSystemFrame::OnPreview(wxCommandEvent& WXUNUSED(event))
{
	if(m_world->GetWorldDefinition().GetSymbols().size() == 0)
	{
		wxMessageBox("You must have at least one symbol in your world definition to preview.", "Error opening preview", wxOK, this);
		return;
	}

	PreviewWindow previewWindow(this, *m_world);
	previewWindow.ShowModal();
}
