#include "ProductionEditDialog.h"

#include "ComposerUtils.h"

BEGIN_EVENT_TABLE(ProductionEditDialog, wxDialog)
	EVT_BUTTON(POS_SET_BUTTON, ProductionEditDialog::OnPosSet)
	EVT_BUTTON(SCALE_SET_BUTTON, ProductionEditDialog::OnScaleSet)
	EVT_BUTTON(ROTATE_RESET_BUTTON, ProductionEditDialog::OnRotateReset)
	EVT_BUTTON(ADD_COMP_BUTTON, ProductionEditDialog::OnAddRSSym)
	EVT_BUTTON(DUP_COMP_BUTTON, ProductionEditDialog::OnDupRSSym)
	EVT_BUTTON(REM_COMP_BUTTON, ProductionEditDialog::OnRemoveRSSym)
	EVT_RADIOBUTTON(PM_APPROX, ProductionEditDialog::HandleApproxPreview)
	EVT_RADIOBUTTON(PM_WIREFRAME, ProductionEditDialog::HandleWireframePreview)
	EVT_RADIOBUTTON(PM_NONE, ProductionEditDialog::HandleNonePreview)
	EVT_CHECKBOX(DRAW_BOUNDS, ProductionEditDialog::HandleDrawBoundsToggle)
	EVT_COMBOBOX(FM_COMBO, ProductionEditDialog::HandleFMChanged)
	EVT_LISTBOX(MY_COMPS_LIST, ProductionEditDialog::OnCompSelectedInList)
END_EVENT_TABLE()

static const char* s_fadeChoices[] =
{
	"Fade In",
	"Fade Out",
	"No Fade",
	"Fade In and Out",
};

ProductionEditDialog::ProductionEditDialog(wxWindow* parent, const wxString& title, ToolWorldDefinition& world, const ToolProduction* prod)
	: wxDialog(parent, -1, title, wxDefaultPosition, wxSize(800, 600), wxDEFAULT_DIALOG_STYLE|wxRESIZE_BORDER|wxMAXIMIZE_BOX)
	, m_world(world)
	, m_prod(prod)
	, m_scene(m_world, m_prod)
{
	new(m_sceneChangedConnection) ProductionEditScene::SceneChangedEvent::Connection(m_scene.GetSceneChangedEventPort(), boost::function<void ()>(boost::bind(&ProductionEditDialog::OnSceneChanged, this)));

	wxBoxSizer* boxSizer = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(boxSizer);

	m_quadView = new QuadViewPanel(m_scene, this);
	boxSizer->Add(m_quadView, 1, wxEXPAND|wxALL, 5);

	wxPanel* toolPanel = new wxPanel(this, -1);
	boxSizer->Add(toolPanel, 0, wxEXPAND|wxALL, 5);

	wxBoxSizer* toolSizer = new wxBoxSizer(wxVERTICAL);
	toolPanel->SetSizer(toolSizer);


	toolSizer->Add(new wxStaticText(toolPanel, -1, wxString("Fade Type:")), 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

	const wxString fadeOptions[] =
	{
		s_fadeChoices[0],
		s_fadeChoices[1],
		s_fadeChoices[2],
		s_fadeChoices[3],
	};
	
	m_fadeTypeBox = new wxComboBox(toolPanel, FM_COMBO, wxString("No Fade"), wxDefaultPosition, wxDefaultSize, 4, fadeOptions, wxCB_READONLY);
	toolSizer->Add(m_fadeTypeBox, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);
	m_fadeTypeBox->SetValue(s_fadeChoices[m_prod->GetProduction()->GetFM()]);
	

	wxStaticBoxSizer* compSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel, wxString("Productions"));
	toolSizer->Add(compSizer, 1, wxEXPAND|wxALL, 5);

	m_componentsList = new wxListBox(toolPanel, MY_COMPS_LIST);
	compSizer->Add(m_componentsList, 1, wxEXPAND|wxALL, 5);

	wxBoxSizer* horizSizer = new wxBoxSizer(wxHORIZONTAL);
	compSizer->Add(horizSizer);

	m_addCompButton = new wxButton(toolPanel, ADD_COMP_BUTTON, wxString("Add"));
	horizSizer->Add(m_addCompButton, 0, wxALIGN_CENTER_VERTICAL, 5);

	m_dupCompButton = new wxButton(toolPanel, DUP_COMP_BUTTON, wxString("Dup"));
	horizSizer->Add(m_dupCompButton, 0, wxALIGN_CENTER_VERTICAL, 5);
	m_dupCompButton->Disable();

	m_remCompButton = new wxButton(toolPanel, REM_COMP_BUTTON, wxString("Remove"));
	horizSizer->Add(m_remCompButton, 0, wxALIGN_CENTER_VERTICAL, 5);
	m_remCompButton->Disable();

	{
		// Attributes
		wxStaticBoxSizer* selectedAttributesSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel);
		toolSizer->Add(selectedAttributesSizer, 0, wxEXPAND|wxALL, 5);
		
		wxBoxSizer* attribsSizer = new wxBoxSizer(wxHORIZONTAL);
		selectedAttributesSizer->Add(attribsSizer, 1, wxEXPAND|wxALL, 5);

		wxStaticBoxSizer* posSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel, "Translation");
		attribsSizer->Add(posSizer, 1, wxEXPAND|wxALL, 2);

		wxFlexGridSizer* posGridSizer = new wxFlexGridSizer(3,2, 0,1);
		posSizer->Add(posGridSizer, 0, wxEXPAND|wxALL, 2);

		wxStaticText* xLabel = new wxStaticText(toolPanel, -1, wxString("X:"));
		posGridSizer->Add(xLabel, 0, wxEXPAND|wxALL, 5);
		m_xPosText = new wxTextCtrl(toolPanel, -1, wxString(""), wxDefaultPosition, wxSize(50, 20));
		posGridSizer->Add(m_xPosText, 0, wxEXPAND|wxALL, 5);

		wxStaticText* yLabel = new wxStaticText(toolPanel, -1, wxString("Y:"));
		posGridSizer->Add(yLabel, 0, wxEXPAND|wxALL, 5);
		m_yPosText = new wxTextCtrl(toolPanel, -1, wxString(""), wxDefaultPosition, wxSize(50, 20));
		posGridSizer->Add(m_yPosText, 0, wxEXPAND|wxALL, 5);

		wxStaticText* zLabel = new wxStaticText(toolPanel, -1, wxString("Z:"));
		posGridSizer->Add(zLabel, 0, wxEXPAND|wxALL, 5);
		m_zPosText = new wxTextCtrl(toolPanel, -1, wxString(""), wxDefaultPosition, wxSize(50, 20));
		posGridSizer->Add(m_zPosText, 0, wxEXPAND|wxALL, 5);
		
		m_setPosButton = new wxButton(toolPanel, POS_SET_BUTTON, wxString("Set"));
		posSizer->Add(m_setPosButton, 0, wxEXPAND|wxALL, 2);

		wxBoxSizer* rightSizer = new wxBoxSizer(wxVERTICAL);
		attribsSizer->Add(rightSizer, 0, wxEXPAND|wxALL, 2);

		wxStaticBoxSizer* scaleSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel, "Scale");
		rightSizer->Add(scaleSizer, 0, wxALL, 2);

		m_scaleText = new wxTextCtrl(toolPanel, -1);
		scaleSizer->Add(m_scaleText, 0, wxEXPAND|wxALL, 5);
		
		m_setScaleButton = new wxButton(toolPanel, SCALE_SET_BUTTON, wxString("Set"));
		scaleSizer->Add(m_setScaleButton, 0, wxEXPAND|wxALL, 5);

		wxStaticBoxSizer* rotationSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel, "Rotation");
		rightSizer->Add(rotationSizer, 0, wxALL, 2);

		m_rotateResetButton = new wxButton(toolPanel, ROTATE_RESET_BUTTON, wxString("Reset"));
		rotationSizer->Add(m_rotateResetButton, 0, wxEXPAND|wxALL, 5);

		UpdateSelectedAttributes();
	}

	wxStaticBoxSizer* viewOptionsSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel, wxString("View Options"));
	toolSizer->Add(viewOptionsSizer, 0, wxEXPAND|wxALL, 5);

	wxStaticBoxSizer* previewModeSizer = new wxStaticBoxSizer(wxVERTICAL, toolPanel, wxString("Preview Mode"));
	viewOptionsSizer->Add(previewModeSizer, 0, wxEXPAND|wxALL, 5);

	m_pmApproximations = new wxRadioButton(toolPanel, PM_APPROX, wxString("Approximations"), wxDefaultPosition, wxDefaultSize, wxRB_GROUP);
	previewModeSizer->Add(m_pmApproximations, 0, wxEXPAND|wxALL, 2);
	m_pmWireframe = new wxRadioButton(toolPanel, PM_WIREFRAME, wxString("Wireframe"));
	previewModeSizer->Add(m_pmWireframe, 0, wxEXPAND|wxALL, 2);
	m_pmNone = new wxRadioButton(toolPanel, PM_NONE, wxString("None"));
	previewModeSizer->Add(m_pmNone, 0, wxEXPAND|wxALL, 2);

	m_drawBounds = new wxCheckBox(toolPanel, DRAW_BOUNDS, wxString("Draw Bounding Boxes"));
	viewOptionsSizer->Add(m_drawBounds, 0, wxEXPAND|wxALL, 5);

	switch(m_scene.GetPreviewMode())
	{
		case ProductionEditScene::PM_APPROXIMATIONS: m_pmApproximations->SetValue(true); break;
		case ProductionEditScene::PM_NONE: m_pmNone->SetValue(true); break;
		default: ASSERT(false);
	};
	m_drawBounds->SetValue(m_scene.GetDrawBounds());

	RecreateComponentsList();

	new(m_prodRSAddedConn) Event<ToolWorldDefinition::ProductionRSAddedSig>::Connection(m_world.GetProductionRSAddedEventPort(), boost::bind(&ProductionEditDialog::HandleProductionRSAdded, this, _1, _2));
	new(m_prodRSRemovedConn) Event<ToolWorldDefinition::ProductionRSRemovedSig>::Connection(m_world.GetProductionRSRemovedEventPort(), boost::bind(&ProductionEditDialog::HandleProductionRSRemoved, this, _1, _2));
	new(m_selectedChangedConn) Event<void ()>::Connection(m_scene.GetSelectionChangedPort(), boost::bind(&ProductionEditDialog::HandleSelectionChanged, this));
}

ProductionEditDialog::~ProductionEditDialog()
{
	m_prodRSAddedConn.Clear();
	m_prodRSRemovedConn.Clear();
	m_selectedChangedConn.Clear();

	delete m_quadView;
}

class AddRSSymbolDialog : public wxDialog
{
	ToolWorldDefinition& m_world;

	wxListBox* m_symList;
	wxButton* m_okButton;
	wxButton* m_cancelButton;

	enum
	{
		SYMBOL_LISTBOX = 400,

	};

public:
	AddRSSymbolDialog(wxWindow* parent, ToolWorldDefinition& world)
		: wxDialog(parent, -1, wxString("Add Symbol to Production"), wxDefaultPosition, wxSize(200, 400))
		, m_world(world)
	{
		wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
		SetSizer(mainSizer);

		wxStaticText* symbolsTxt = new wxStaticText(this, -1, wxString("Symbols:"));
		mainSizer->Add(symbolsTxt, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

		m_symList = new wxListBox(this, SYMBOL_LISTBOX);
		mainSizer->Add(m_symList, 1, wxEXPAND|wxALL, 5);

		// Populate the list box with all the available symbols
		std::list<const ToolSymbol*> symList = m_world.GetSymbolList();
		for(std::list<const ToolSymbol*>::const_iterator iter = symList.begin(); iter != symList.end(); ++iter)
		{
			m_symList->Append((*iter)->GetName().c_str());
		}

		wxBoxSizer* buttonSizer = new wxBoxSizer(wxHORIZONTAL);
		mainSizer->Add(buttonSizer, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

		m_okButton = new wxButton(this, wxID_OK, wxString("Okay"));
		buttonSizer->Add(m_okButton, 1, wxEXPAND|wxALL, 5);
		m_okButton->Disable();

		m_cancelButton = new wxButton(this, wxID_CANCEL, wxString("Cancel"));
		buttonSizer->Add(m_cancelButton, 1, wxEXPAND|wxALL, 5);
		
	};

	const ToolSymbol* GetRSSym()
	{
		ASSERT(m_symList->GetSelection() != wxNOT_FOUND);

		return m_world.GetSymbol(m_symList->GetString(m_symList->GetSelection()).c_str());
	}

	void OnSymbolSelected(wxCommandEvent&)
	{
		if(m_symList->GetSelection() != wxNOT_FOUND)
		{
			m_okButton->Enable();
		}
	}

	DECLARE_EVENT_TABLE()
};
BEGIN_EVENT_TABLE(AddRSSymbolDialog, wxDialog)
	EVT_LISTBOX(SYMBOL_LISTBOX, AddRSSymbolDialog::OnSymbolSelected)
END_EVENT_TABLE()

void ProductionEditDialog::OnAddRSSym(wxCommandEvent&)
{
	AddRSSymbolDialog addRSSym(this, m_world);

	if(addRSSym.ShowModal() == wxID_OK)
	{
		m_world.AddSymbolToProductionRS(m_prod, addRSSym.GetRSSym());
	}
}

void ProductionEditDialog::OnDupRSSym(wxCommandEvent&)
{
	// Duping a production right side involves creating a new right side with the same symbol
	// as selected, and then copying the transformation over
	ASSERT(m_scene.GetSelectedIndex() != -1);

	const Symbol* symbol = m_prod->GetProduction()->GetChildren()[m_scene.GetSelectedIndex()]->sym;
	const ToolSymbol* toolSym = m_world.GetSymbol(symbol);

	m_world.AddSymbolToProductionRS(m_prod, toolSym);

	const CompositionTransformation& formerTrans = m_prod->GetTransformation(m_scene.GetSelectedIndex());
	
	int newProdIndex = m_prod->GetProduction()->GetChildren().size() - 1;

	m_world.EditProductionComp(m_prod, newProdIndex, formerTrans);
}

void ProductionEditDialog::OnRemoveRSSym(wxCommandEvent&)
{
	if(m_scene.GetSelectedIndex() != -1)
	{
		m_world.RemoveSymbolFromProductionRS(m_prod, m_scene.GetSelectedIndex());
	}
}

void ProductionEditDialog::HandleProductionRSAdded(const ToolProduction* prod, int index)
{
	if(prod == m_prod)
	{
		m_componentsList->Append(m_world.GetStringForSymbol(prod->GetProduction()->GetChildren()[index]->sym).c_str());
	}
}

void ProductionEditDialog::HandleProductionRSRemoved(const ToolProduction* prod, int index)
{
	if(prod == m_prod)
	{
		m_componentsList->Delete(index);

		if(m_scene.GetSelectedIndex() == m_prod->GetProduction()->GetChildren().size())
		{
			if(m_scene.GetSelectedIndex() == 0) m_scene.SetSelected(-1);
			else m_scene.SetSelected(m_scene.GetSelectedIndex() - 1);
		}
		else if(m_scene.GetSelectedIndex() != -1)
		{
			ASSERT(m_scene.GetSelectedIndex() < m_prod->GetProduction()->GetChildren().size());
			m_componentsList->SetSelection(m_scene.GetSelectedIndex());

			m_scene.RecreateScene();
		}
	}
}


void ProductionEditDialog::OnSceneChanged()
{
	UpdateSelectedAttributes();
}

void ProductionEditDialog::UpdateSelectedAttributes()
{
	if(m_scene.GetSelectedIndex() == -1)
	{
		m_xPosText->Disable();
		m_yPosText->Disable();
		m_zPosText->Disable();
		m_setPosButton->Disable();

		m_scaleText->Disable();
		m_setScaleButton->Disable();
		
		m_rotateResetButton->Disable();
	}
	else
	{
		m_xPosText->Enable();
		m_yPosText->Enable();
		m_zPosText->Enable();
		m_setPosButton->Enable();

		m_scaleText->Enable();
		m_setScaleButton->Enable();

		const CompositionTransformation& selectedTrans = m_prod->GetTransformation(m_scene.GetSelectedIndex());

		{
			std::ostringstream str;
			str << selectedTrans.GetPosition()[0];
			m_xPosText->SetValue(wxString(str.str()));
		}
		{
			std::ostringstream str;
			str << selectedTrans.GetPosition()[1];
			m_yPosText->SetValue(wxString(str.str()));
		}
		{
			std::ostringstream str;
			str << selectedTrans.GetPosition()[2];
			m_zPosText->SetValue(wxString(str.str()));
		}
		{
			std::ostringstream str;
			str << selectedTrans.GetScale();
			m_scaleText->SetValue(wxString(str.str()));
		}

		m_rotateResetButton->Enable();
	}
}

void ProductionEditDialog::RecreateComponentsList()
{
	m_componentsList->Clear();

	const Production::ChildList& children =  m_prod->GetProduction()->GetChildren();
	for(Production::ChildList::const_iterator iter = children.begin(); iter != children.end(); ++iter)
	{
		m_componentsList->Append(m_world.GetStringForSymbol((*iter)->sym).c_str());
	}
}

void ProductionEditDialog::HandleApproxPreview(wxCommandEvent&)
{
	m_scene.SetPreviewMode(ProductionEditScene::PM_APPROXIMATIONS);
}

void ProductionEditDialog::HandleWireframePreview(wxCommandEvent&)
{
	m_scene.SetPreviewMode(ProductionEditScene::PM_WIREFRAME);
}

void ProductionEditDialog::HandleNonePreview(wxCommandEvent&)
{
	m_scene.SetPreviewMode(ProductionEditScene::PM_NONE);
}

	
void ProductionEditDialog::HandleDrawBoundsToggle(wxCommandEvent&)
{
	m_scene.SetDrawBounds(m_drawBounds->GetValue());
}

void ProductionEditDialog::HandleFMChanged(wxCommandEvent&)
{
	for(int i=0; i<Production::FM_NUM_FADEMODES; ++i)
	{
		if(m_fadeTypeBox->GetValue() == s_fadeChoices[i])
		{
			m_world.EditProductionFadeMode(m_prod, static_cast<Production::FadeMode>(i));
			return;
		}
	}
	
	ASSERT(false);
}

void ProductionEditDialog::OnCompSelectedInList(wxCommandEvent&)
{
	m_scene.SetSelected((m_componentsList->GetSelection() != wxNOT_FOUND ? m_componentsList->GetSelection() : -1));
}

void ProductionEditDialog::HandleSelectionChanged()
{
	int newSelection = (m_scene.GetSelectedIndex() == -1 ? wxNOT_FOUND : m_scene.GetSelectedIndex());

	if(newSelection != m_componentsList->GetSelection())
	{
		m_componentsList->SetSelection(newSelection);
	}

	if(m_scene.GetSelectedIndex() == -1)
	{
		m_dupCompButton->Disable();
		m_remCompButton->Disable();
	}
	else
	{
		m_dupCompButton->Enable();
		m_remCompButton->Enable();
	}

	UpdateSelectedAttributes();
}

void ProductionEditDialog::OnPosSet(wxCommandEvent&)
{
	if(ValidateFloatString(m_xPosText->GetValue()) && ValidateFloatString(m_yPosText->GetValue()) && ValidateFloatString(m_zPosText->GetValue()))
	{
		// Okay, we got some valid numbers here, time to make some shit happen
		float toX, toY, toZ;

		{
			std::istringstream in(m_xPosText->GetValue().c_str());
			in >> toX;
		}
		{
			std::istringstream in(m_yPosText->GetValue().c_str());
			in >> toY;
		}
		{
			std::istringstream in(m_zPosText->GetValue().c_str());
			in >> toZ;
		}

		ASSERT(m_scene.GetSelectedIndex() != -1);
		const CompositionTransformation& selectedTrans = m_prod->GetTransformation(m_scene.GetSelectedIndex());

		// Look at the current transformation to get the delta between that and the desired transformation
		float dx = toX - selectedTrans.GetPosition()[0];
		float dy = toY - selectedTrans.GetPosition()[1];
		float dz = toZ - selectedTrans.GetPosition()[2];

		// And now attempt to apply this.
		m_scene.MoveSelected(Vector<4>(dx, dy, dz, 1));
	}
	else
	{
		::wxMessageBox(wxString("Position fields must be a string of decimal digits with at most one period.  Ie: 1, 3., .5, 03.5"), wxString("Invalid Input"), wxOK|wxICON_ERROR);
	}
}

void ProductionEditDialog::OnScaleSet(wxCommandEvent&)
{
	if(ValidateFloatString(m_scaleText->GetValue()))
	{
		float toScale;
		std::istringstream in(m_scaleText->GetValue().c_str());
		in >> toScale;

		ASSERT(m_scene.GetSelectedIndex() != -1);
		const CompositionTransformation& selectedTrans = m_prod->GetTransformation(m_scene.GetSelectedIndex());

		float deltaScale = toScale / selectedTrans.GetScale();

		m_scene.ScaleSelected(deltaScale);
	}
	else
	{
		::wxMessageBox(wxString("Scale field must be a string of decimal digits with at most one period.  Ie: 1, 3., .5, 03.5"), wxString("Invalid Input"), wxOK|wxICON_ERROR);
	}
}

void ProductionEditDialog::OnRotateReset(wxCommandEvent&)
{
	ASSERT(m_scene.GetSelectedIndex() != -1);
	const CompositionTransformation& selectedTrans = m_prod->GetTransformation(m_scene.GetSelectedIndex());

	UnitQuaternion deltaRot = UnitQuaternion() * selectedTrans.GetRotation().Inverse();

	m_scene.RotateSelected(deltaRot);
}
