#include "ProductionEditScene.h"
#include "R3Utils.h"


ProductionEditScene::ProductionEditScene(ToolWorldDefinition& world, const ToolProduction* prod)
: m_world(world)
, m_prod(prod)
, m_selected(-1)
, m_previewMode(PM_APPROXIMATIONS)
, m_drawBounds(true)
, m_showMarker(false)
, m_curFrame(0)
{
	RecreateScene();

	m_prodRSAddedConn = new Event<void (const ToolProduction*, int)>::Connection(m_world.GetProductionRSAddedEventPort(), boost::bind(&ProductionEditScene::HandleProductionRSAdded, this, _1, _2));
	m_prodChangedConn = new Event<void (const ToolProduction*)>::Connection(m_world.GetProductionChangedEventPort(), boost::bind(&ProductionEditScene::HandleProductionChanged, this, _1));
}

ProductionEditScene::~ProductionEditScene()
{
	delete m_prodChangedConn;
	delete m_prodRSAddedConn;
}

ProductionEditScene::ObjectFrameAssets::ObjectFrameAssets()
{	
	white.ambientColour = Vector<4>(1,1,1,1);
	red.ambientColour = Vector<4>(1,0,0,1);
	green.ambientColour = Vector<4>(0,1,0,1);
	blue.ambientColour = Vector<4>(0,0,1,1);
	selectCol.ambientColour = Vector<4>(1,0,1,1);

	x.a = Vector<4>(0,0,0,1);
	x.b = Vector<4>(1,0,0,1);

	y.a = Vector<4>(0,0,0,1);
	y.b = Vector<4>(0,1,0,1);

	z.a = Vector<4>(0,0,0,1);
	z.b = Vector<4>(0,0,1,1);

	others[0].a = Vector<4>(1,0,0,1);
	others[0].b = Vector<4>(1,0,1,1);
	others[1].a = Vector<4>(1,0,1,1);
	others[1].b = Vector<4>(0,0,1,1);

	others[2].a = Vector<4>(1,0,0,1);
	others[2].b = Vector<4>(1,1,0,1);
	others[3].a = Vector<4>(1,0,1,1);
	others[3].b = Vector<4>(1,1,1,1);
	others[4].a = Vector<4>(0,0,1,1);
	others[4].b = Vector<4>(0,1,1,1);

	others[5].a = Vector<4>(0,1,0,1);
	others[5].b = Vector<4>(1,1,0,1);
	others[6].a = Vector<4>(1,1,0,1);
	others[6].b = Vector<4>(1,1,1,1);
	others[7].a = Vector<4>(1,1,1,1);
	others[7].b = Vector<4>(0,1,1,1);
	others[8].a = Vector<4>(0,1,1,1);
	others[8].b = Vector<4>(0,1,0,1);
}

void ProductionEditScene::ObjectFrameAssets::AddToFrame(GraphicsFrame& toFrame, int modelView, bool selected)
{
	toFrame.AddPacket(GraphicsFrame::Packet(modelView, &x, &(selected ? selectCol : red), 2));
	toFrame.AddPacket(GraphicsFrame::Packet(modelView, &y, &(selected ? selectCol : green), 2));
	toFrame.AddPacket(GraphicsFrame::Packet(modelView, &z, &(selected ? selectCol : blue), 2));

	for(int i=0; i<9; ++i)
	{
		toFrame.AddPacket(GraphicsFrame::Packet(modelView, &others[i], &(selected ? selectCol : white), 1.5f));
	}
}

void ProductionEditScene::RenderScene(GraphicsContext& context, const GraphicsContext::RenderOptions& options)
{
	context.RenderGraphicsFrame(*m_frames[m_curFrame], options);
}

void ProductionEditScene::AddSceneHelpers(GraphicsFrame& toFrame)
{
	int identMat = toFrame.RegisterMatrix(Matrix<4,4>(Matrix<4,4>::Identity));

	m_objectFrameAssets.AddToFrame(toFrame, identMat);

	if(m_showMarker)
	{
		int markerMat = toFrame.RegisterMatrix(MakeTranslationMatrix(m_markerPos.Trim()));
		m_objectFrameAssets.AddToFrame(toFrame, markerMat);
	}
}

void ProductionEditScene::HandleProductionRSAdded(const ToolProduction* prod, int index)
{
	ASSERT(prod == m_prod);

	RecreateScene();
}

void ProductionEditScene::HandleProductionChanged(const ToolProduction* prod)
{
	ASSERT(prod == m_prod);

	RecreateScene();
}

void ProductionEditScene::RecreateScene()
{
	m_curFrame = (m_curFrame + 1) % NUM_FRAMES;

	m_frames[m_curFrame] = std::auto_ptr<GraphicsFrame>(new GraphicsFrame(1000, 1000));
	GraphicsFrame& curFrame = *m_frames[m_curFrame];

	curFrame.SetBackgroundColour(Vector<3>(0.6f, 0.6f, 0.6f));

	AddSceneHelpers(curFrame);

	if(m_drawBounds)
	{
		const Production::ChildList& children =  m_prod->GetProduction()->GetChildren();
		for(Production::ChildList::const_iterator iter = children.begin(); iter != children.end(); ++iter)
		{
			int prodRSSymMat = curFrame.RegisterMatrix(Matrix<4,4>((*iter)->transform));
			m_objectFrameAssets.AddToFrame(curFrame, prodRSSymMat, (m_selected != -1 && *iter == GetSelectedDesc()));
		}
	}

	switch(m_previewMode)
	{
		case PM_APPROXIMATIONS:
			DoPreviewModeApproximations(curFrame, false);
			break;
		case PM_WIREFRAME:
			DoPreviewModeApproximations(curFrame, true);
			break;
		case PM_NONE:
			break;
		default: ASSERT(false);
	};

	curFrame.Lock();

	m_sceneChangedEvent.Fire();
}

void ProductionEditScene::DoPreviewModeApproximations(GraphicsFrame& curFrame, bool wireframe)
{
	const Production::ChildList& children =  m_prod->GetProduction()->GetChildren();

	// Now draw the approximations to all the symbols in the production
	for(Production::ChildList::const_iterator iter = children.begin(); iter != children.end(); ++iter)
	{
		if(!(*iter)->sym->GetApproxRenderFunc().empty())
		{
			int prodRSSymMat = curFrame.RegisterMatrix(Matrix<4,4>((*iter)->transform));
			(*iter)->sym->GetApproxRenderFunc()(&curFrame, &((*iter)->transform), 1, 0, wireframe);
		}
	}
}

void ProductionEditScene::PlaceMarker(const Vector<4>& pos)
{
	m_showMarker = true;
	m_markerPos = pos;

	RecreateScene();
}

static bool RayIntersectUnitBox(const Vector<3>& rayStart, const Vector<3>& dir)
{
	for(int side=0; side<2; ++side)
	{
		for(int i=0; i<3; ++i)
		{
			float t = (side - rayStart[i]) / dir[i];

			int j;
			for(j=0; j<3; ++j)
			{
				if(i==j) continue;

				float comp = rayStart[j] + t * dir[j];

				if(comp < 0 || comp > 1) break;
			}
			if(j != 3) continue; // No intersection

			// There was an intersection
			return true;
		}
	}

	return false;
}

void ProductionEditScene::TrySelectItem(const Vector<4>& rayStart, const Vector<4>& otherPointOnRay)
{
	int oldSelected = m_selected;

	bool found = false;
	int index = 0;
	const Production::ChildList& prodChildren = m_prod->GetProduction()->GetChildren();
	for(Production::ChildList::const_iterator iter = prodChildren.begin(); iter != prodChildren.end(); ++iter, ++index)
	{
		const Production::ChildDescriptor* desc = *iter;

		if(m_selected == index) continue;

		// Transform the ray points in to object space, and then do the ray test in there
		if(RayIntersectUnitBox((desc->invTransform*rayStart).Trim(), (desc->invTransform*(otherPointOnRay-rayStart)).Trim()))
		{
			// Select this one then
			PrivateSetSelected(index);
			found = true;
			break;
		}
	}

	if(!found) PrivateSetSelected(-1);

	if(oldSelected != m_selected)
	{
		RecreateScene();
	}
}

void ProductionEditScene::PrivateSetSelected(int selected)
{
	ASSERT(selected >= -1);
	ASSERT(selected < (int)m_prod->GetProduction()->GetChildren().size());

	if(m_selected != selected)
	{
		m_selected = selected;

		m_selectedChangedEvent.Fire();
	}
}

void ProductionEditScene::SetSelected(int selected)
{
	PrivateSetSelected(selected);
	RecreateScene();
}



// This represents the concept of a modification to a CompositionTransformation
// object.  For example, we might want to adjust the position of a CompositionTransformation,
// this would be accomplished through creating a TransformationModifier, and applying that
// to the CompositionTransformation
class TransformModifier
{
public:
	virtual void ApplyModifier(CompositionTransformation& applyTrans, float lambda) const = 0;
};

class TranslationTransformModifier : public TransformModifier
{
	int m_axis;
	float m_moveAmount;

public:
	TranslationTransformModifier(int axis, float moveAmount)
		: m_axis(axis)
		, m_moveAmount(moveAmount)
	{};

	void ApplyModifier(CompositionTransformation& applyTrans, float lambda) const
	{
		ASSERT(lambda >= 0 && lambda <= 1);

		Vector<3> moveVec(0,0,0);
		moveVec[m_axis] = m_moveAmount * lambda;

		applyTrans.SetPosition(applyTrans.GetPosition() + moveVec);
	}
};

class UniScaleTransformModifier : public TransformModifier
{
	float m_scaleFactor;

public:
	UniScaleTransformModifier(float scaleFactor)
		: m_scaleFactor(scaleFactor)
	{};

	void ApplyModifier(CompositionTransformation& applyTrans, float lambda) const
	{
		ASSERT(lambda >= 0 && lambda <= 1);

		float scaleAmount = 1 + (m_scaleFactor - 1) * lambda;

		applyTrans.SetScale(applyTrans.GetScale() * scaleAmount);
	}
};

class RotationTransformModifier : public TransformModifier
{
	UnitQuaternion m_rot;

public:
	RotationTransformModifier(const UnitQuaternion& rot)
		: m_rot(rot)
	{};

	void ApplyModifier(CompositionTransformation& applyTrans, float lambda) const
	{
		ASSERT(lambda >= 0 && lambda <= 1);

		float newW = cos(acos(m_rot.GetW()) * lambda);
		if(newW == 1) return;

		float a = (1 - newW*newW) / (m_rot.GetV().LengthSquared());
		UnitQuaternion processedRot(m_rot.GetV() * a);

		applyTrans.SetRotation(processedRot * applyTrans.GetRotation());
	}
};

static void MakeTransformedCubePoints(const Matrix<4,4>& trans, Vector<4> cubePoints[8])
{
	cubePoints[0] = trans.GetColumn(3); // (0,0,0)
	cubePoints[1] = (trans.GetColumn(0) + trans.GetColumn(3)); // (1,0,0)
	cubePoints[2] = (cubePoints[1] + trans.GetColumn(1)); // (1,1,0)
	cubePoints[3] = (trans.GetColumn(1) + trans.GetColumn(3)); // (0,1,0)
	cubePoints[4] = (cubePoints[3] + trans.GetColumn(2)); // (0,1,1)
	cubePoints[5] = (trans.GetColumn(2) + trans.GetColumn(3)); // (0,0,1)
	cubePoints[6] = (trans.GetColumn(0) + cubePoints[5]); // (1,0,1)
	cubePoints[7] = (trans.GetColumn(1) + cubePoints[6]); // (1,1,1)
}

// I could almost name this "IsContraction" except
// this test is stricter than a contraction
static bool IsWithinUnit(const Matrix<4,4>& trans)
{
	// Check that the transformed cube points remain within the unit cube
	Vector<4> cubePoints[8];
	MakeTransformedCubePoints(trans, cubePoints);
	for(int i=0; i<8; ++i)
	{
		for(int j=0; j<3; ++j)
		{
			if(cubePoints[i][j] < 0 || cubePoints[i][j] > 1) return false;
		}
	}
	return true;
}

// Takes a transformation and a transformation modifier and returns a new CompositionTransformation that has the modifier applied to it but still has all points in the unit cube
static CompositionTransformation ValidateTransformByIteration(const CompositionTransformation& origTransform, const TransformModifier& modifier)
{
	// The algorithm here works by "trying" a transformation out, and seeing if the result is invalid
	// or not.  If it is invalid, then we try applying "half of the transformation" and checking the result
	// of that.  We iterate on this a constant number of times and finally return the best fit transformation.
	// This allows us to restrict the transformations any way we like, without having to solve how individual
	// transformations should be modified to result in a new transformation that isn't invalid.  Unfortunately,
	// this way is less exact though.  Looks like it's exact enough.
	float lastGoodAmount = 0.0f;

	float amount = 1.0f;
	float amountDelta = 1.0f;

	const int ITERATE_ON_TRANSFORM = 16; // How many times do we halve a transform to get it close to the limit
	for(int i=0; i<ITERATE_ON_TRANSFORM; ++i)
	{
		amountDelta *= 0.5f;

		CompositionTransformation nextTransform(origTransform);
		modifier.ApplyModifier(nextTransform, amount);

		if(IsWithinUnit(nextTransform.ToMatrix()))
		{
			if(i == 0) return nextTransform; // No need for iteration if the transformation is valid

			lastGoodAmount = amount;
			amount += amountDelta;
		}
		else
		{
			// We broke the rules, go back a bit
			amount -= amountDelta;
		}
	}

	// Return the best transform within the cube
	CompositionTransformation nextTransform(origTransform);
	modifier.ApplyModifier(nextTransform, lastGoodAmount);
	return nextTransform;
}

void ProductionEditScene::MoveSelected(const Vector<4>& moveVec)
{
	ASSERT(m_selected != -1);

	CompositionTransformation finalMove(m_prod->GetTransformation(m_selected));
	// Amalgamate[take that Ryan] the individual coordinate transformations differences in to one final transformation
	for(int i=0; i<3; ++i)
	{
		CompositionTransformation nextMove = ValidateTransformByIteration(m_prod->GetTransformation(m_selected), TranslationTransformModifier(i, moveVec[i]));
		Vector<3> nextPos(finalMove.GetPosition());
		nextPos[i] = nextMove.GetPosition()[i];

		finalMove.SetPosition(nextPos);
	}


	ASSERT(IsWithinUnit(finalMove.ToMatrix()));

	// Pass on the move vector to the world definition
	m_world.EditProductionComp(m_prod, m_selected, finalMove);
}

void ProductionEditScene::ScaleSelected(float scale)
{
	ASSERT(m_selected != -1);
	ASSERT(scale > 0.0f);

	m_world.EditProductionComp(m_prod, m_selected, ValidateTransformByIteration(m_prod->GetTransformation(m_selected), UniScaleTransformModifier(scale)));
}

void ProductionEditScene::RotateSelected(const UnitQuaternion& rot)
{
	ASSERT(m_selected != -1);

	m_world.EditProductionComp(m_prod, m_selected, ValidateTransformByIteration(m_prod->GetTransformation(m_selected), RotationTransformModifier(rot)));
}

void ProductionEditScene::SetPreviewMode(PreviewMode mode)
{
	if(mode == m_previewMode) return;

	m_previewMode = mode;

	RecreateScene();
}

void ProductionEditScene::SetDrawBounds(bool drawBounds)
{
	if(m_drawBounds == drawBounds) return;

	m_drawBounds = drawBounds;

	RecreateScene();
}

