#include "QuadViewPanel.h"

#include "ProductionEditScene.h"

#include <Projection.h>
#include <ArcBall.h>

#include <boost/bind.hpp>

BEGIN_EVENT_TABLE(EditPanel, wxGLCanvas)
	EVT_MOTION(EditPanel::OnMouseMoved)
	EVT_LEFT_DOWN(EditPanel::OnMouseDown)
	EVT_LEFT_DCLICK(EditPanel::OnMouseDoubleClick)
	EVT_LEFT_UP(EditPanel::OnMouseReleased)
	EVT_RIGHT_DOWN(EditPanel::OnRightDown)
	EVT_RIGHT_UP(EditPanel::OnRightUp)
	EVT_MOUSEWHEEL(EditPanel::OnMouseWheel)
	EVT_SIZE(EditPanel::OnResize)
	EVT_MOUSE_CAPTURE_LOST(EditPanel::OnMouseCaptureLost)
	EVT_PAINT(EditPanel::OnPaint)
	EVT_ERASE_BACKGROUND(EditPanel::OnEraseBackground)
END_EVENT_TABLE()



BEGIN_EVENT_TABLE(QuadViewPanel, wxPanel)
	EVT_ACTIVATE(QuadViewPanel::OnActivate)
END_EVENT_TABLE()



QuadViewPanel::QuadViewPanel(ProductionEditScene& scene, wxWindow* parent)
: wxPanel(parent, -1, wxDefaultPosition, wxDefaultSize, wxFULL_REPAINT_ON_RESIZE)
, m_mainSizer(0)
, m_editPanelHolder(0)
, m_editsGrid(0)
, m_scene(scene)
{
	for(int i=0; i<4; ++i) m_editPanels[i] = 0;

	m_mainSizer = new wxBoxSizer(wxHORIZONTAL);
	SetSizer(m_mainSizer);

	m_editPanelHolder = new wxPanel(this, -1);

	m_mainSizer->Add(m_editPanelHolder, 1, wxEXPAND, 3);

	m_editsGrid = new wxGridSizer(2,2, 3,3);
	m_editPanelHolder->SetSizer(m_editsGrid);

	for(int i=0; i<4; ++i)
	{
		m_editPanels[i] = new EditPanel(m_scene, m_editPanelHolder, wxPoint(-1, -1), wxSize(-1,-1));
		m_editPanels[i]->SetBackgroundColour(wxColour(255,0,0));

		switch(i)
		{
			case 0: m_editPanels[i]->SetDefaultViewRotation(UnitQuaternion(0.0053110905, -0.47563329, 0.0021694824, -0.87961638)); break;
			case 1: m_editPanels[i]->SetDefaultViewRotation(UnitQuaternion(M_SQRT1_2,0,0,M_SQRT1_2)); break;
			case 2: m_editPanels[i]->SetDefaultViewRotation(UnitQuaternion(0,-M_SQRT1_2,0,M_SQRT1_2)); break;
			case 3: m_editPanels[i]->SetDefaultViewRotation(UnitQuaternion(0,-M_SQRT1_2,0,M_SQRT1_2)*UnitQuaternion(0,-M_SQRT1_2,0,M_SQRT1_2)); break;
		};

		m_editPanels[i]->ResetPanelView();

		m_editsGrid->Add(m_editPanels[i], 1, wxEXPAND);
	}

	// Put the axis-aligned views in to orthographic mode
	m_editPanels[1]->SetProjectionType(EditPanel::PT_ORTHOGRAPHIC);
	m_editPanels[2]->SetProjectionType(EditPanel::PT_ORTHOGRAPHIC);
	m_editPanels[3]->SetProjectionType(EditPanel::PT_ORTHOGRAPHIC);

	// Lock control of the view for the auxilary views
	m_editPanels[1]->SetUserControllableView(false);
	m_editPanels[2]->SetUserControllableView(false);
	m_editPanels[3]->SetUserControllableView(false);


	m_editsGrid->Layout();
}

QuadViewPanel::~QuadViewPanel()
{
	for(int i=0; i<4; ++i)
	{
		delete m_editPanels[i];
	}
}

void QuadViewPanel::OnActivate(wxActivateEvent& event)
{
	event.Skip();

	m_editsGrid->Layout();
}


EditPanel::EditPanel(ProductionEditScene& scene, wxWindow* parent, const wxPoint& pos, const wxSize& size)
: wxGLCanvas(parent, -1, 0, pos, size, wxSUNKEN_BORDER|wxCLIP_CHILDREN|wxFULL_REPAINT_ON_RESIZE)
, m_context(0)
, m_fov(45.0f)
, m_panning(false)
, m_rotating(false)
, m_movingSelected(false)
, m_scalingSelected(false)
, m_rotatingSelected(false)
, m_orthoZoom(1.0f)
, m_scene(scene)
, m_projType(PT_PERSPECTIVE)
, m_orthoViewRad(0.6f)
, m_userControllableView(true)
{
	m_context = new wxWidgetGraphicsContext(*this);

	m_camera.SetFocusPoint(Vector<4>(0.5f,0.5f,0.5f,1));
	m_defaultCamera.SetFocusPoint(Vector<4>(0.5f,0.5f,0.5f,1));

	m_sceneChangedConnection = new ProductionEditScene::SceneChangedEvent::Connection(scene.GetSceneChangedEventPort(), boost::function<void ()>(boost::bind(&EditPanel::DoRedraw, this)));
}

EditPanel::~EditPanel()
{
	delete m_sceneChangedConnection;
	delete m_context;
}

void EditPanel::OnMouseMoved(wxMouseEvent& event)
{
	m_mouseMoved = true;

	m_prevMouseX = m_curMouseX;
	m_prevMouseY = m_curMouseY;
	m_curMouseX = event.GetX();
	m_curMouseY = event.GetY();

	if(m_panning)
	{
		int dx = m_curMouseX - m_prevMouseX;
		int dy = m_curMouseY - m_prevMouseY;

		wxSize winSize = GetClientSize();

		float moveX = (0.5f * dx)/ winSize.GetWidth();
		float moveY = (-0.5f * dy) / winSize.GetHeight();

		// We now have the measure of how much we should pan by, we now
		// determine the plane to pan on, from the rotation matrix
		Matrix<4,4> rotMat = m_camera.GetRotation().ToMatrix();

		Vector<4> focusPoint = m_camera.GetFocusPoint();
		focusPoint += -rotMat.GetColumn(0) * moveX;
		focusPoint += -rotMat.GetColumn(1) * moveY;

		m_camera.SetFocusPoint(focusPoint);

		m_viewChanged.Fire();
		
		Refresh();
	}

	if(m_rotating && GetUserControllableView())
	{
		int width;
		int height;
		GetClientSize(&width, &height);
		float minDim = (width < height ? width : height);

		float centeredPrevX = m_prevMouseX - (width/2);
		float centeredPrevY = m_prevMouseY - (height/2);
		float centeredCurrX = m_curMouseX - (width/2);
		float centeredCurrY = m_curMouseY - (height/2);

		Vector<2> from(2.0f * centeredPrevX/minDim, -(2.0f * centeredPrevY/minDim));
		Vector<2> to(2.0f * centeredCurrX/minDim, -(2.0f * centeredCurrY/minDim));

		UnitQuaternion rotQuat = GetRotationFromArcBallPointDelta(from, to);
		m_camera.SetRotation(m_camera.GetRotation()*rotQuat.Conjugate());

		m_viewChanged.Fire();
		
		Refresh();
	}

	if(m_movingSelected && m_scene.HaveSelection())
	{
		int dx = m_curMouseX - m_prevMouseX;
		int dy = m_curMouseY - m_prevMouseY;

		wxSize winSize = GetClientSize();

		float moveX = (0.5f * dx)/ winSize.GetWidth();
		float moveY = (-0.5f * dy) / winSize.GetHeight();

		// We now have the measure of how much we should pan by, we now
		// determine the plane to pan on, from the rotation matrix
		Matrix<4,4> rotMat = m_camera.GetRotation().ToMatrix();

		m_scene.MoveSelected(rotMat.GetColumn(0) * moveX + rotMat.GetColumn(1) * moveY);
	}

	if(m_scalingSelected && m_scene.HaveSelection())
	{
		// Compute our "distance from center" delta, per coordinate
		wxSize winSize = GetClientSize();

		int curDistX = abs(m_curMouseX - winSize.GetWidth()/2);
		int curDistY = abs(m_curMouseY - winSize.GetHeight()/2);

		int prevDistX = abs(m_prevMouseX - winSize.GetWidth()/2);
		int prevDistY = abs(m_prevMouseY - winSize.GetHeight()/2);

		int dx = curDistX - prevDistX;
		int dy = -(curDistY - prevDistY);

		float moveX = (0.5f * dx)/ winSize.GetWidth();
		float moveY = (-0.5f * dy) / winSize.GetHeight();

		float scaleDegree = moveX + moveY;

		float scale = pow(2, scaleDegree*2);
		
		// We now have the measure of how much we should scale by
		// based off of whether the user is moving the mouse away from
		// the center, or towards it.
		m_scene.ScaleSelected(scale);
	}

	if(m_rotatingSelected && m_scene.HaveSelection())
	{
		int width;
		int height;
		GetClientSize(&width, &height);
		float minDim = (width < height ? width : height);

		float centeredPrevX = m_prevMouseX - (width/2);
		float centeredPrevY = m_prevMouseY - (height/2);
		float centeredCurrX = m_curMouseX - (width/2);
		float centeredCurrY = m_curMouseY - (height/2);

		Vector<2> from(2.0f * centeredPrevX/minDim, -(2.0f * centeredPrevY/minDim));
		Vector<2> to(2.0f * centeredCurrX/minDim, -(2.0f * centeredCurrY/minDim));

		UnitQuaternion rotQuat = GetRotationFromArcBallPointDelta(from, to);

		// Now put this rotation relative to our current view angle
		Vector<4> rotDir(rotQuat.GetV()[0], rotQuat.GetV()[1], rotQuat.GetV()[2], 0);
		rotDir = m_camera.GetRotation().ToMatrix() * rotDir;
		
		UnitQuaternion processedRot(rotDir.Trim(), rotQuat.GetW());

		m_scene.RotateSelected(processedRot);
	}
}

void EditPanel::SetViewRotation(const UnitQuaternion& rot)
{
	m_camera.SetRotation(rot);

	m_viewChanged.Fire();
	
	Refresh();
}

void EditPanel::SetDefaultViewRotation(const UnitQuaternion& rot)
{
	m_defaultCamera.SetRotation(rot);
}

void EditPanel::OnMouseDown(wxMouseEvent& event)
{
	event.Skip();

	if(!HasCapture())
	{
		CaptureMouse();
		m_mouseMoved = false;

		if(event.ShiftDown())
		{
			m_panning = true;
		}
		else
		{
			m_rotating = true;
		}

		m_prevMouseX = event.GetX();
		m_prevMouseY = event.GetY();
		m_curMouseX = event.GetX();
		m_curMouseY = event.GetY();
	}
}

void EditPanel::OnMouseReleased(wxMouseEvent& event)
{
	event.Skip();

	if(HasCapture())
	{
		ReleaseMouse();
	}

	m_panning = false;
	m_rotating = false;

	if(!m_mouseMoved)
	{
		TrySelectItem(m_curMouseX, m_curMouseY);
	}
}

void EditPanel::OnRightDown(wxMouseEvent& event)
{
	event.Skip();

	if(!HasCapture())
	{
		CaptureMouse();
		m_mouseMoved = false;

		if(event.ShiftDown())
		{
			m_movingSelected = true;
		}
		else if(event.ControlDown())
		{
			m_scalingSelected = true;
		}
		else
		{
			m_rotatingSelected = true;
		}
	}
}

void EditPanel::OnRightUp(wxMouseEvent& event)
{
	event.Skip();

	if(HasCapture())
	{
		ReleaseMouse();
	}

	m_movingSelected = false;
	m_scalingSelected = false;
	m_rotatingSelected = false;
}


void EditPanel::OnMouseWheel(wxMouseEvent& event)
{
	event.Skip();

	int rotate = event.GetWheelRotation();

	if(m_projType == PT_ORTHOGRAPHIC)
	{
		m_orthoZoom = m_orthoZoom * pow(0.9f, -(float)event.GetWheelRotation()/120);

		const float MAX_ORTHO_ZOOM = 64.0f;
		const float MIN_ORTHO_ZOOM = 1.0f;
		if(m_orthoZoom > MAX_ORTHO_ZOOM)
		{
			m_orthoZoom = MAX_ORTHO_ZOOM;
		}

		if(m_orthoZoom < MIN_ORTHO_ZOOM)
		{
			m_orthoZoom = MIN_ORTHO_ZOOM;
		}
	}
	else
	{
		m_camera.SetDistance(m_camera.GetDistance() * pow(0.9f, (float)event.GetWheelRotation()/120));
		
		const float MIN_CAMERA_DISTANCE = 0.1f;
		const float MAX_CAMERA_DISTANCE = 50.0f;

		if(m_camera.GetDistance() < MIN_CAMERA_DISTANCE)
		{
			m_camera.SetDistance(MIN_CAMERA_DISTANCE);
		}
		if(m_camera.GetDistance() > MAX_CAMERA_DISTANCE)
		{
			m_camera.SetDistance(MAX_CAMERA_DISTANCE);
		}
	}

	m_viewChanged.Fire();
	
	Refresh();
}


void EditPanel::OnMouseCaptureLost(wxMouseCaptureLostEvent& event)
{
	m_panning = false;
	m_rotating = false;
}


void EditPanel::OnResize(wxSizeEvent& event)
{
	event.Skip();

	RefreshContextSize();

	m_viewChanged.Fire();
	
	Refresh();
}

void EditPanel::OnEraseBackground(wxEraseEvent& event)
{
	// No fucking background.  Why the hell would wxGLCanvas have a background by default?
}

void EditPanel::OnPaint(wxPaintEvent& event)
{
	DoRedraw();
	event.Skip();
}

void EditPanel::DoRedraw()
{
	Matrix<4,4> projMatrix = GetProjectionMatrix();

	GraphicsContext::RenderOptions options(projMatrix*m_camera.GetModelView());

	m_scene.RenderScene(GetContext(), options);
}

void EditPanel::RefreshContextSize()
{
	int width;
	int height;
	GetClientSize(&width, &height);

	m_context->ResizeScreen(width, height);
}

void EditPanel::TrySelectItem(int x, int y)
{
	// Create our ray as if there was no camera transform,
	// then apply the inverse camera transform to that and use the
	// resulting ray to do our intersections with

	int width;
	int height;
	GetClientSize(&width, &height);
	float ratio = (float)width/height;

	float normX = ((float)x/width) * 2 - 1.0f;
	float normY = ((float)(height-y)/height) * 2 - 1.0f;

	float t = tan(M_PI * m_fov / 360.0f);
	
	float z = -1;

	// Transform the origin to get the camera position in world space
	Vector<4> camPoint(0,0,0,1);
	Vector<4> viewPoint;	
	switch(m_projType)
	{
		case PT_PERSPECTIVE:
		{
			viewPoint = Vector<4>(-z*normX * ratio * t, -z*normY * t, z, 1);
		} break;

		case PT_ORTHOGRAPHIC:
		{
			camPoint = Vector<4>(normX * m_orthoViewRad, normY * m_orthoViewRad, 0, 1);
			viewPoint = Vector<4>(normX * m_orthoViewRad, normY * m_orthoViewRad, z, 1);
		} break;
	};

	

	// worldPoint here is the point 1 unit of distance infront of the camera, on the "click line"
	Vector<4> worldPoint(m_camera.GetModelView().Inverse()*viewPoint);
	Vector<4> camOriginWorld(m_camera.GetModelView().Inverse()*camPoint);

	m_scene.TrySelectItem(camOriginWorld, worldPoint);
}

Matrix<4,4> EditPanel::GetProjectionMatrix() const
{
	int width;
	int height;
	GetClientSize(&width, &height);

	switch(m_projType)
	{
		case PT_PERSPECTIVE:
		{
			return MakePerspectiveProjectionMatrix(m_fov, (float)width/height, 0.001f, 100.0f);
		} break;

		case PT_ORTHOGRAPHIC:
		{
			float invZoom = 1/m_orthoZoom;
			return MakeOrthographicProjectionMatrix(-m_orthoViewRad*invZoom, m_orthoViewRad*invZoom, -m_orthoViewRad*invZoom, m_orthoViewRad*invZoom, 0.001f, 100.0f);
		} break;

		default: ASSERT(false);
  	};

	// dummy return value
	return Matrix<4,4>();
}

void EditPanel::OnMouseDoubleClick(wxMouseEvent& event)
{
	ResetPanelView();
}

void EditPanel::ResetPanelView()
{
	m_orthoZoom = 1.0f;
	m_camera = m_defaultCamera;
	m_viewChanged.Fire();
	Refresh();
}
