#include "SceneExplorer.h"

#include "WorldDefinition.h"
#include "ToolWorldDefinition.h"

#include <ScopedFStream.h>
#include <ThreadPrimitives.h>

#include <OSWindow.h>

#include <iostream>

static Mutex s_quitMutex;
static Condition s_quitCond;
static bool s_quitFlag = false;

static PotentialObject<OSWindow> s_window;

static void SignalQuitCond()
{
	Mutex::ScopedLock lock(s_quitMutex);

	s_quitFlag = true;

	s_quitCond.SignalAll();
}

static void HandleKeyEvent(const KeyboardInput::KeyEvent& evt)
{
	if(evt.pressed)
	{
		switch(evt.key)
		{
			case KeyboardInput::KT_F:
			{
				if(s_window->GetWindowMode() == OSWindow::M_WINDOWED)
				{
					s_window->SetWindowMode(OSWindow::M_FULLSCREEN);
				}
				else
				{
					s_window->SetWindowMode(OSWindow::M_WINDOWED);
				}
			} break;

			case KeyboardInput::KT_ESC:
			{
				SignalQuitCond();
			} break;
		};
	}
}

int main(int argc, char** args)
{
	try
	{
		if(argc < 2)
		{
			std::cerr << "Pass the Linfinity World file (*.lin) that you would like to explore" << std::endl;
			return -1;
		}
		{
			// Make sure that the file exists (if it doesn't it will throw an exception			
			ScopedFStream<std::ifstream> ifs(args[1]);
		}
		
		std::cout << "Loading '" << args[1] << "'" << std::endl;
		
		
		FileFetcher fetcher(10000000);
		ResourceManager resMan(g_standardAllocator, 100000000, fetcher);

		// Create and load our definitions
		ToolWorldDefinition worldDef(resMan, args[1]);

		// Create the window that will host our app
		new(s_window) OSWindow("Scene Explorer", 1024, 768);
		Event<void ()>::Connection quitEventConn(s_window->GetQuitEventPort(), &SignalQuitCond);
		KeyboardInput::KeyListenerConnection keyConn(s_window->GetKeyboardInput(), &HandleKeyEvent);

		// Create the scene explorer to explore our world
		SceneExplorer sceneExplorer(worldDef.GetWorldDefinition(), s_window->GetGraphicsContext(), s_window->GetKeyboardInput(), s_window->GetMouseInput());

		// And now wait for it to quit
		{
			Mutex::ScopedLock lock(s_quitMutex);

			while(!s_quitFlag)
			{
				s_quitCond.Wait(s_quitMutex);
			}
		}
	}
	catch(std::exception& e)
	{
		std::cerr << "FATAL ERROR: " << e.what() << std::endl;
	}

	return 0;
}
